from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import (
    Container,
    Mapping,
    Sequence,
)
from decimal import Decimal
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
    Self,
    TypeAlias,
    cast,
)
from uuid import UUID, uuid4

import sqlalchemy as sa
import trafaret as t
from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError
from sqlalchemy import CheckConstraint
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncSession
from sqlalchemy.orm import contains_eager, foreign, relationship, selectinload
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.common.config import model_definition_iv
from ai.backend.common.types import (
    AccessKey,
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    EndpointId,
    RuntimeVariant,
    SessionTypes,
    VFolderID,
    VFolderMount,
    VFolderUsageMode,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.creator import DeploymentCreator
from ai.backend.manager.data.deployment.scale import (
    AutoScalingAction,
    AutoScalingCondition,
    AutoScalingRule,
    AutoScalingRuleCreator,
)
from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    DeploymentMetadata,
    DeploymentNetworkSpec,
    DeploymentState,
    ExecutionSpec,
    ModelRevisionSpec,
    MountMetadata,
    ReplicaSpec,
    ResourceSpec,
)
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.session.types import SessionStatus

from ..config.loader.legacy_etcd_loader import LegacyEtcdLoader
from ..data.model_serving.creator import EndpointCreator
from ..data.model_serving.types import (
    EndpointAutoScalingRuleData,
    EndpointData,
    EndpointLifecycle,
    EndpointTokenData,
)
from ..errors.api import InvalidAPIParameters
from ..errors.common import ObjectNotFound, ServiceUnavailable
from ..models.storage import StorageSessionManager
from ..types import MountOptionModel, UserScope
from .base import (
    GUID,
    Base,
    DecimalType,
    EndpointIDColumn,
    EnumValueType,
    IDColumn,
    ResourceSlotColumn,
    StrEnumType,
    StructuredJSONObjectListColumn,
    URLColumn,
)
from .image import ImageRow
from .routing import RouteStatus
from .scaling_group import scaling_groups
from .user import UserRow
from .vfolder import prepare_vfolder_mounts

if TYPE_CHECKING:
    from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient

    from .gql import GraphQueryContext

__all__ = (
    "EndpointRow",
    "ModelServiceHelper",
    "EndpointStatistics",
    "EndpointTokenRow",
    "EndpointAutoScalingRuleRow",
    "EndpointLifecycle",
)


ModelServiceSerializableConnectionInfo: TypeAlias = dict[str, list[dict[str, Any]]]

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


class EndpointRow(Base):
    __tablename__ = "endpoints"

    __table_args__ = (
        CheckConstraint(
            sa.or_(
                sa.column("lifecycle_stage") == EndpointLifecycle.DESTROYED.value,
                sa.column("image").isnot(None),
            ),
            name="ck_image_required_unless_destroyed",
        ),
        sa.Index(
            "ix_endpoints_unique_name_when_not_destroyed",
            "name",
            "domain",
            "project",
            unique=True,
            postgresql_where=(sa.column("lifecycle_stage") != EndpointLifecycle.DESTROYED.value),
        ),
    )

    id = EndpointIDColumn()
    name = sa.Column("name", sa.String(length=512), nullable=False)
    created_user = sa.Column("created_user", GUID, nullable=False)
    session_owner = sa.Column("session_owner", GUID, nullable=False)
    # minus session count means this endpoint is requested for removal
    replicas = sa.Column("replicas", sa.Integer, nullable=False, default=0, server_default="0")
    desired_replicas = sa.Column(
        "desired_replicas", sa.Integer, nullable=True, default=None, server_default=sa.null()
    )
    image = sa.Column("image", GUID)
    model = sa.Column(
        "model",
        GUID,
        sa.ForeignKey("vfolders.id", ondelete="SET NULL"),
        nullable=True,
    )
    model_mount_destination = sa.Column(
        "model_mount_destination",
        sa.String(length=1024),
        nullable=False,
        default="/models",
        server_default="/models",
    )
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="RESTRICT"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="RESTRICT"),
        nullable=False,
    )
    resource_group = sa.Column(
        "resource_group",
        sa.ForeignKey("scaling_groups.name", ondelete="RESTRICT"),
        index=True,
        nullable=False,
    )
    lifecycle_stage = sa.Column(
        "lifecycle_stage",
        EnumValueType(EndpointLifecycle),
        nullable=False,
        default=EndpointLifecycle.PENDING,
    )
    tag = sa.Column("tag", sa.String(length=64), nullable=True)
    startup_command = sa.Column("startup_command", sa.Text, nullable=True)
    bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
    callback_url = sa.Column("callback_url", URLColumn, nullable=True, default=sa.null())
    environ = sa.Column("environ", pgsql.JSONB(), nullable=True, default={})
    open_to_public = sa.Column("open_to_public", sa.Boolean, default=False)
    runtime_variant = sa.Column(
        "runtime_variant",
        StrEnumType(RuntimeVariant),
        nullable=False,
        default=RuntimeVariant.CUSTOM,
    )

    model_definition_path = sa.Column("model_definition_path", sa.String(length=128), nullable=True)

    resource_slots = sa.Column("resource_slots", ResourceSlotColumn(), nullable=False)
    url = sa.Column("url", sa.String(length=1024))
    resource_opts = sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={})
    cluster_mode = sa.Column(
        "cluster_mode",
        sa.String(length=16),
        nullable=False,
        default=ClusterMode.SINGLE_NODE,
        server_default=ClusterMode.SINGLE_NODE.name,
    )
    cluster_size = sa.Column(
        "cluster_size", sa.Integer, nullable=False, default=1, server_default="1"
    )

    extra_mounts = sa.Column(
        "extra_mounts",
        StructuredJSONObjectListColumn(VFolderMount),
        nullable=False,
        default=[],
        server_default="[]",
    )

    retries = sa.Column("retries", sa.Integer, nullable=False, default=0, server_default="0")
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    destroyed_at = sa.Column(
        "destroyed_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    routings = relationship("RoutingRow", back_populates="endpoint_row")
    tokens = relationship(
        "EndpointTokenRow",
        back_populates="endpoint_row",
        primaryjoin="foreign(EndpointTokenRow.endpoint) == EndpointRow.id",
    )
    endpoint_auto_scaling_rules = relationship(
        "EndpointAutoScalingRuleRow", back_populates="endpoint_row"
    )
    image_row = relationship(
        "ImageRow",
        primaryjoin=lambda: foreign(EndpointRow.image) == ImageRow.id,
        foreign_keys=[image],
        back_populates="endpoints",
    )

    model_row = relationship("VFolderRow", back_populates="endpoints")
    created_user_row = relationship(
        "UserRow",
        back_populates="created_endpoints",
        foreign_keys=[created_user],
        primaryjoin=lambda: foreign(EndpointRow.created_user) == UserRow.uuid,
    )
    session_owner_row = relationship(
        "UserRow",
        back_populates="owned_endpoints",
        foreign_keys=[session_owner],
        primaryjoin=lambda: foreign(EndpointRow.session_owner) == UserRow.uuid,
    )

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_tokens: bool = False,
        load_image: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        load_model: bool = False,
    ) -> Self:
        """
        :raises: sqlalchemy.orm.exc.NoResultFound
        """
        query = sa.select(EndpointRow).filter(EndpointRow.id == endpoint_id)
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(
                selectinload(EndpointRow.image_row).selectinload(ImageRow.aliases)
            )
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if load_model:
            query = query.options(selectinload(EndpointRow.model_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if row is None:
            raise NoResultFound
        return row

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> list[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(EndpointRow.lifecycle_stage.in_(status_filter))
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def batch_load(
        cls,
        session: AsyncSession,
        endpoint_ids: Sequence[EndpointId],
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_(status_filter) & EndpointRow.id.in_(endpoint_ids)
            )
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def list_by_model(
        cls,
        session: AsyncSession,
        model_id: UUID,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_(status_filter) & (EndpointRow.model == model_id)
            )
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    async def create_auto_scaling_rule(
        self,
        session: AsyncSession,
        metric_source: AutoScalingMetricSource,
        metric_name: str,
        threshold: Decimal,
        comparator: AutoScalingMetricComparator,
        step_size: int,
        cooldown_seconds: int = 300,
        min_replicas: int | None = None,
        max_replicas: int | None = None,
    ) -> EndpointAutoScalingRuleRow:
        row = EndpointAutoScalingRuleRow(
            id=uuid4(),
            endpoint=self.id,
            metric_source=metric_source,
            metric_name=metric_name,
            threshold=threshold,
            comparator=comparator,
            step_size=step_size,
            cooldown_seconds=cooldown_seconds,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )
        session.add(row)
        return row

    @property
    def terminatable_route_statuses(self) -> set[RouteStatus]:
        if self.lifecycle_stage == EndpointLifecycle.DESTROYING:
            return {
                RouteStatus.PROVISIONING,
                RouteStatus.HEALTHY,
                RouteStatus.UNHEALTHY,
                RouteStatus.FAILED_TO_START,
            }
        else:
            return {
                RouteStatus.HEALTHY,
                RouteStatus.UNHEALTHY,
                RouteStatus.FAILED_TO_START,
            }

    @staticmethod
    async def delegate_endpoint_ownership(
        db_session: AsyncSession,
        owner_user_uuid: UUID,
        target_user_uuid: UUID,
        target_access_key: AccessKey,
    ) -> None:
        from .routing import RoutingRow
        from .session import KernelLoadingStrategy, SessionRow

        endpoint_rows = await EndpointRow.list(
            db_session,
            user_uuid=owner_user_uuid,
            load_session_owner=True,
            load_routes=True,
            load_tokens=True,
        )
        session_ids: List[UUID] = []
        for row in endpoint_rows:
            row.session_owner = target_user_uuid
            for token_row in cast(list[EndpointTokenRow], row.tokens):
                token_row.delegate_ownership(target_user_uuid)
            for routing_row in cast(list[RoutingRow], row.routings):
                routing_row.delegate_ownership(target_user_uuid)
                session_ids.append(routing_row.session)
        session_rows = await SessionRow.list_sessions(
            db_session, session_ids, kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS
        )
        for session_row in session_rows:
            session_row.delegate_ownership(target_user_uuid, target_access_key)

    async def generate_route_info(
        self, db_sess: AsyncSession
    ) -> ModelServiceSerializableConnectionInfo:
        from .kernel import KernelRow
        from .routing import RoutingRow

        active_routes = await RoutingRow.list(db_sess, self.id, load_session=True)
        running_main_kernels = await KernelRow.batch_load_main_kernels_by_session_id(
            db_sess,
            [
                r.session
                for r in active_routes
                if r.status in RouteStatus.active_route_statuses()
                and r.session
                and r.session_row.status in [SessionStatus.RUNNING, SessionStatus.CREATING]
            ],
        )
        if (num_routes_without_session := len(active_routes) - len(running_main_kernels)) > 0:
            log.info(
                "generate_route_info(): There are {} active routes without corresponding RUNNING sessions, "
                "which may be still provisioning or being terminated. (Endpoint: {})",
                num_routes_without_session,
                self.id,
            )
        session_id_to_route_map = {r.session: r for r in active_routes}
        connection_info: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)
        for kernel in running_main_kernels:
            if kernel.service_ports is None:
                log.debug(
                    "generate_route_info(): Kernel {} has no service ports defined. Skipping.",
                    kernel.id,
                )
                continue
            num_inference_ports = len([*filter(lambda x: x["is_inference"], kernel.service_ports)])
            if num_inference_ports > 1:
                log.warning(
                    "generate_route_info(): Multiple ({}) inference ports found. "
                    "Currently only the first-seen inference port is used. (Endpoint: {})",
                    num_inference_ports,
                    self.id,
                )
            for port_info in kernel.service_ports:
                if port_info["is_inference"]:
                    connection_info[port_info["name"]].append({
                        "session_id": str(kernel.session_id),
                        "route_id": str(session_id_to_route_map[kernel.session_id].id),
                        "kernel_host": kernel.kernel_host,
                        "kernel_port": port_info["host_ports"][0],
                    })
                    break
        return connection_info

    def to_data(self) -> EndpointData:
        return EndpointData(
            id=self.id,
            name=self.name,
            image=self.image_row.to_dataclass() if self.image_row else None,
            domain=self.domain,
            project=self.project,
            resource_group=self.resource_group,
            resource_slots=self.resource_slots,
            url=self.url,
            model=self.model,
            model_definition_path=self.model_definition_path,
            model_mount_destination=self.model_mount_destination,
            created_user_id=self.created_user,
            created_user_email=(
                self.created_user_row.email if self.created_user_row is not None else None
            ),
            session_owner_id=self.session_owner,
            session_owner_email=self.session_owner_row.email if self.session_owner_row else "",
            tag=self.tag,
            startup_command=self.startup_command,
            bootstrap_script=self.bootstrap_script,
            callback_url=self.callback_url,
            environ=self.environ,
            resource_opts=self.resource_opts,
            replicas=self.replicas,
            cluster_mode=ClusterMode(self.cluster_mode),
            cluster_size=self.cluster_size,
            open_to_public=self.open_to_public,
            created_at=self.created_at,
            destroyed_at=self.destroyed_at,
            retries=self.retries,
            lifecycle_stage=self.lifecycle_stage,
            runtime_variant=self.runtime_variant,
            extra_mounts=self.extra_mounts,
            routings=[routing.to_data() for routing in self.routings] if self.routings else None,
        )

    @classmethod
    def from_creator(cls, creator: EndpointCreator) -> Self:
        """
        Create an EndpointRow instance from an EndpointCreator instance.
        """
        return cls(
            name=creator.name,
            model_definition_path=creator.model_definition_path,
            created_user=creator.created_user,
            session_owner=creator.session_owner,
            replicas=creator.replicas,
            image=creator.image,
            model=creator.model,
            domain=creator.domain,
            project=creator.project,
            resource_group=creator.resource_group,
            resource_slots=creator.resource_slots,
            cluster_mode=creator.cluster_mode,
            cluster_size=creator.cluster_size,
            extra_mounts=creator.extra_mounts,
            runtime_variant=creator.runtime_variant,
            model_mount_destination=creator.model_mount_destination,
            tag=creator.tag,
            startup_command=creator.startup_command,
            bootstrap_script=creator.bootstrap_script,
            callback_url=creator.callback_url,
            environ=creator.environ,
            resource_opts=creator.resource_opts,
            open_to_public=creator.open_to_public,
        )

    @classmethod
    async def from_deployment_creator(
        cls,
        db_session: AsyncSession,
        creator: DeploymentCreator,
    ) -> Self:
        """
        Create an EndpointRow instance from a DeploymentCreator instance.

        Args:
            db_session: Database session for resolving image information
            creator: DeploymentCreator containing deployment configuration

        Returns:
            EndpointRow instance with image ID resolved from ImageIdentifier

        Raises:
            InvalidAPIParameters: If image is not specified in creator
            ImageNotFound: If image cannot be resolved
        """
        # Image is required
        if not creator.model_revision.image_identifier:
            raise InvalidAPIParameters("Image must be specified in DeploymentCreator")

        # Resolve image ID from ImageIdentifier
        image_row = await ImageRow.lookup(
            db_session,
            creator.model_revision.image_identifier,
        )

        return cls(
            name=creator.metadata.name,
            created_user=creator.metadata.created_user,
            session_owner=creator.metadata.session_owner,
            replicas=creator.replica_spec.replica_count,
            image=image_row.id,
            model=creator.model_revision.mounts.model_vfolder_id,
            model_mount_destination=creator.model_revision.mounts.model_mount_destination,
            model_definition_path=creator.model_revision.mounts.model_definition_path,
            domain=creator.metadata.domain,
            project=creator.metadata.project,
            resource_group=creator.metadata.resource_group,
            tag=creator.metadata.tag,
            startup_command=creator.model_revision.execution.startup_command,
            bootstrap_script=creator.model_revision.execution.bootstrap_script,
            callback_url=creator.model_revision.execution.callback_url,
            environ=creator.model_revision.execution.environ,
            open_to_public=creator.network.open_to_public,
            runtime_variant=creator.model_revision.execution.runtime_variant,
            resource_slots=creator.model_revision.resource_spec.resource_slots,
            url=creator.network.url,
            resource_opts=creator.model_revision.resource_spec.resource_opts,
            cluster_mode=creator.model_revision.resource_spec.cluster_mode,
            cluster_size=creator.model_revision.resource_spec.cluster_size,
            extra_mounts=creator.model_revision.mounts.extra_mounts,
            # Fields not in creator - use defaults
            lifecycle_stage=EndpointLifecycle.PENDING,
            retries=0,
        )

    def to_deployment_info(self) -> DeploymentInfo:
        """
        Convert EndpointRow to DeploymentInfo dataclass.

        If image_row is loaded (via selectinload), ImageIdentifier will be populated.
        Otherwise, ImageIdentifier will be None.
        """
        # Create ImageIdentifier only if image_row is loaded
        image_identifier = ImageIdentifier(
            canonical=self.image_row.name,
            architecture=self.image_row.architecture,
        )
        return DeploymentInfo(
            id=self.id,
            metadata=DeploymentMetadata(
                name=self.name,
                domain=self.domain,
                project=self.project,
                resource_group=self.resource_group,
                created_user=self.created_user,
                session_owner=self.session_owner,
                created_at=self.created_at,
                tag=self.tag,
            ),
            state=DeploymentState(
                lifecycle=self.lifecycle_stage,
                retry_count=self.retries,
            ),
            replica_spec=ReplicaSpec(
                replica_count=self.replicas,
                desired_replica_count=self.desired_replicas,
            ),
            network=DeploymentNetworkSpec(
                open_to_public=self.open_to_public,
                url=self.url,
            ),
            model_revisions=[
                ModelRevisionSpec(
                    image_identifier=image_identifier,
                    resource_spec=ResourceSpec(
                        cluster_mode=self.cluster_mode,
                        cluster_size=self.cluster_size,
                        resource_slots=self.resource_slots,
                        resource_opts=self.resource_opts,
                    ),
                    mounts=MountMetadata(
                        model_vfolder_id=self.model,
                        model_definition_path=self.model_definition_path,
                        model_mount_destination=self.model_mount_destination,
                        extra_mounts=self.extra_mounts,
                    ),
                    execution=ExecutionSpec(
                        startup_command=self.startup_command,
                        bootstrap_script=self.bootstrap_script,
                        environ=self.environ,
                        runtime_variant=self.runtime_variant,
                        callback_url=self.callback_url,
                    ),
                ),
            ],
        )


class EndpointTokenRow(Base):
    __tablename__ = "endpoint_tokens"

    id = IDColumn()
    token = sa.Column("token", sa.String(), nullable=False)
    endpoint = sa.Column("endpoint", GUID, nullable=True)
    session_owner = sa.Column("session_owner", GUID, nullable=False)
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="CASCADE"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="CASCADE"),
        nullable=False,
    )
    created_at = sa.Column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
    )

    endpoint_row = relationship(
        "EndpointRow",
        back_populates="tokens",
        foreign_keys=[endpoint],
        primaryjoin=lambda: foreign(EndpointTokenRow.endpoint) == EndpointRow.id,
    )

    def __init__(
        self,
        id: UUID,
        token: str,
        endpoint: UUID,
        domain: str,
        project: UUID,
        session_owner: UUID,
    ) -> None:
        self.id = id
        self.token = token
        self.endpoint = endpoint
        self.domain = domain
        self.project = project
        self.session_owner = session_owner

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        *,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_endpoint: bool = False,
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointTokenRow)
            .filter(EndpointTokenRow.endpoint == endpoint_id)
            .order_by(sa.desc(EndpointTokenRow.created_at))
        )
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        token: str,
        *,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_endpoint: bool = False,
    ) -> Self:
        query = sa.select(EndpointTokenRow).filter(EndpointTokenRow.token == token)
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if not row:
            raise NoResultFound
        return row

    def delegate_ownership(self, user_uuid: UUID) -> None:
        self.session_owner = user_uuid

    def to_dataclass(self) -> EndpointTokenData:
        return EndpointTokenData(
            id=self.id,
            token=self.token,
            endpoint=self.endpoint,
            domain=self.domain,
            project=self.project,
            session_owner=self.session_owner,
            created_at=self.created_at,
        )


class EndpointAutoScalingRuleRow(Base):
    __tablename__ = "endpoint_auto_scaling_rules"

    id = IDColumn()
    metric_source = sa.Column(
        "metric_source", StrEnumType(AutoScalingMetricSource, use_name=False), nullable=False
    )
    metric_name = sa.Column("metric_name", sa.Text(), nullable=False)
    threshold = sa.Column("threshold", DecimalType(), nullable=False)
    comparator = sa.Column(
        "comparator", StrEnumType(AutoScalingMetricComparator, use_name=False), nullable=False
    )
    step_size = sa.Column("step_size", sa.Integer(), nullable=False)
    cooldown_seconds = sa.Column("cooldown_seconds", sa.Integer(), nullable=False, default=300)

    min_replicas = sa.Column("min_replicas", sa.Integer(), nullable=True)
    max_replicas = sa.Column("max_replicas", sa.Integer(), nullable=True)

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    last_triggered_at = sa.Column(
        "last_triggered_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    endpoint = sa.Column(
        "endpoint",
        GUID,
        sa.ForeignKey("endpoints.id", ondelete="CASCADE"),
        nullable=False,
    )

    endpoint_row = relationship(
        "EndpointRow", back_populates="endpoint_auto_scaling_rules", lazy="joined"
    )

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        endpoint_status_filter: Container[EndpointLifecycle] = frozenset([
            EndpointLifecycle.CREATED
        ]),
    ) -> Sequence[Self]:
        query = sa.select(EndpointAutoScalingRuleRow)
        if endpoint_status_filter:
            query = (
                query.join(EndpointAutoScalingRuleRow.endpoint_row)
                .filter(EndpointRow.lifecycle_stage.in_(endpoint_status_filter))
                .options(contains_eager(EndpointAutoScalingRuleRow.endpoint_row))
            )
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def get(
        cls, session: AsyncSession, id: UUID, load_endpoint: bool = False
    ) -> "EndpointAutoScalingRuleRow":
        query = sa.select(EndpointAutoScalingRuleRow).filter(EndpointAutoScalingRuleRow.id == id)
        if load_endpoint:
            query = query.options(selectinload(EndpointAutoScalingRuleRow.endpoint_row))
        result = await session.execute(query)
        row = result.scalar()
        if not row:
            raise ObjectNotFound(object_name="Endpoint Autoscaling Rule")
        return row

    async def remove_rule(
        self,
        session: AsyncSession,
    ) -> None:
        await session.delete(self)

    def to_data(self) -> EndpointAutoScalingRuleData:
        return EndpointAutoScalingRuleData(
            id=self.id,
            metric_source=self.metric_source,
            metric_name=self.metric_name,
            threshold=self.threshold,
            comparator=self.comparator,
            step_size=self.step_size,
            cooldown_seconds=self.cooldown_seconds,
            min_replicas=self.min_replicas,
            max_replicas=self.max_replicas,
            created_at=self.created_at,
            last_triggered_at=self.last_triggered_at,
            endpoint=self.endpoint,
        )

    @classmethod
    def from_creator(cls, endpoint_id: UUID, creator: AutoScalingRuleCreator) -> Self:
        return cls(
            id=uuid4(),
            endpoint=endpoint_id,
            metric_source=creator.condition.metric_source,
            metric_name=creator.condition.metric_name,
            threshold=creator.condition.threshold,
            comparator=creator.condition.comparator,
            step_size=creator.action.step_size,
            cooldown_seconds=creator.action.cooldown_seconds,
            min_replicas=creator.action.min_replicas,
            max_replicas=creator.action.max_replicas,
        )

    def to_autoscaling_rule(self) -> AutoScalingRule:
        return AutoScalingRule(
            id=self.id,
            condition=AutoScalingCondition(
                metric_source=self.metric_source,
                metric_name=self.metric_name,
                threshold=self.threshold,
                comparator=self.comparator,
            ),
            action=AutoScalingAction(
                step_size=self.step_size,
                cooldown_seconds=self.cooldown_seconds,
                min_replicas=self.min_replicas,
                max_replicas=self.max_replicas,
            ),
            created_at=self.created_at,
            last_triggered_at=self.last_triggered_at,
        )


class ModelServiceHelper:
    @staticmethod
    async def check_scaling_group(
        conn: AsyncConnection,
        scaling_group: str,
        owner_access_key: AccessKey,
        target_domain: str,
        target_project: str | UUID,
    ) -> str:
        """
        Wrapper of `registry.check_scaling_group()` with additional guards flavored for
        model service included
        """
        from ai.backend.manager.registry import check_scaling_group

        checked_scaling_group = await check_scaling_group(
            conn,
            scaling_group,
            SessionTypes.INFERENCE,
            owner_access_key,
            target_domain,
            target_project,
        )

        query = (
            sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == checked_scaling_group))
        )

        result = await conn.execute(query)
        sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        if not wsproxy_addr:
            raise ServiceUnavailable("No coordinator configured for this resource group")

        if not sgroup["wsproxy_api_token"]:
            raise ServiceUnavailable("Scaling group not ready to start model service")

        return checked_scaling_group

    @staticmethod
    async def check_extra_mounts(
        conn: AsyncConnection,
        legacy_etcd_loader: LegacyEtcdLoader,
        storage_manager: StorageSessionManager,
        model_id: UUID,
        model_mount_destination: str,
        extra_mounts: dict[UUID, MountOptionModel],
        user_scope: UserScope,
        resource_policy: dict[str, Any],
    ) -> Sequence[VFolderMount]:
        """
        check if user is allowed to access every folders eagering to mount (other than model VFolder)
        on general session creation lifecycle this check will be completed by `enqueue_session()` function,
        which is not covered by the validation procedure (`create_session(dry_run=True)` call at the bottom part of `create()` API)
        so we have to manually cover this part here.
        """
        if model_id in extra_mounts:
            raise InvalidAPIParameters(
                "Same VFolder appears on both model specification and VFolder mount"
            )

        requested_mounts = [*extra_mounts.keys()]
        requested_mount_map: dict[str | UUID, str] = {
            folder_id: options.mount_destination
            for folder_id, options in extra_mounts.items()
            if options.mount_destination
        }
        requested_mount_options: dict[str | UUID, Any] = {
            folder_id: {
                "type": options.type,
                "permission": options.permission,
            }
            for folder_id, options in extra_mounts.items()
        }
        log.debug(
            "requested mounts: {}, mount_map: {}, mount_options: {}",
            requested_mounts,
            requested_mount_map,
            requested_mount_options,
        )
        allowed_vfolder_types = await legacy_etcd_loader.get_vfolder_types()
        vfolder_mounts = await prepare_vfolder_mounts(
            conn,
            storage_manager,
            allowed_vfolder_types,
            user_scope,
            resource_policy,
            requested_mounts,
            requested_mount_map,
            requested_mount_options,
        )

        for vfolder in vfolder_mounts:
            if vfolder.kernel_path == model_mount_destination:
                raise InvalidAPIParameters(
                    "extra_mounts.mount_destination conflicts with model_mount_destination config. Make sure not to shadow value defined at model_mount_destination as a mount destination of extra VFolders."
                )
            if vfolder.usage_mode == VFolderUsageMode.MODEL:
                raise InvalidAPIParameters(
                    "MODEL type VFolders cannot be added as a part of extra_mounts folder"
                )

        return vfolder_mounts

    @staticmethod
    async def _listdir(
        storage_manager: StorageSessionManager,
        proxy_name: str,
        volume_name: str,
        vfid: VFolderID,
        relpath: str,
    ) -> dict[str, Any]:
        manager_facing_client = storage_manager.get_manager_facing_client(proxy_name)
        result = await manager_facing_client.list_files(
            volume_name,
            str(vfid),
            relpath,
        )
        return cast(dict[str, Any], result)

    @staticmethod
    async def validate_model_definition_file_exists(
        storage_manager: StorageSessionManager,
        folder_host: str,
        vfid: VFolderID,
        suggested_path: str | None,
    ) -> str:
        """
        Checks if model definition file exists in target model VFolder. Returns path to resolved model definition filename.
        Since model service counts both `model-definition.yml` and `model-definition.yaml` as valid definition file name, this function ensures
        at least one model definition file exists under the target VFolder and returns the matched filename.
        """
        proxy_name, volume_name = storage_manager.get_proxy_and_volume(folder_host)

        if suggested_path:
            path = Path(suggested_path)
            storage_reply = await ModelServiceHelper._listdir(
                storage_manager, proxy_name, volume_name, vfid, path.parent.as_posix()
            )
            for item in storage_reply["items"]:
                if item["name"] == path.name:
                    return suggested_path
            else:
                raise InvalidAPIParameters(
                    f"Model definition YAML file {suggested_path} not found inside the model storage"
                )
        else:
            storage_reply = await ModelServiceHelper._listdir(
                storage_manager, proxy_name, volume_name, vfid, "."
            )
            model_definition_candidates = ["model-definition.yaml", "model-definition.yml"]
            for item in storage_reply["items"]:
                if item["name"] in model_definition_candidates:
                    return item["name"]
            else:
                raise InvalidAPIParameters(
                    'Model definition YAML file "model-definition.yaml" or "model-definition.yml" not found inside the model storage'
                )

    @staticmethod
    async def _read_model_definition(
        storage_manager: StorageSessionManager,
        folder_host: str,
        vfid: VFolderID,
        model_definition_filename: str,
    ) -> dict[str, Any]:
        """
        Reads specified model definition file from target VFolder and returns
        """
        proxy_name, volume_name = storage_manager.get_proxy_and_volume(folder_host)
        manager_facing_client = storage_manager.get_manager_facing_client(proxy_name)
        chunks = await manager_facing_client.fetch_file_content(
            volume_name,
            str(vfid),
            f"./{model_definition_filename}",
        )
        model_definition_yaml = chunks.decode("utf-8")
        yaml = YAML()
        return yaml.load(model_definition_yaml)

    @staticmethod
    async def validate_model_definition(
        storage_manager: StorageSessionManager,
        folder_host: str,
        vfid: VFolderID,
        model_definition_path: str,
    ) -> dict[str, Any]:
        """
        Checks if model definition YAML exists and is syntactically perfect.
        Returns validated model definition configuration.
        """
        raw_model_definition = await ModelServiceHelper._read_model_definition(
            storage_manager,
            folder_host,
            vfid,
            model_definition_path,
        )

        try:
            model_definition = model_definition_iv.check(raw_model_definition)
            assert model_definition is not None
            return model_definition
        except t.DataError as e:
            raise InvalidAPIParameters(
                f"Failed to validate model definition from VFolder (ID {vfid.folder_id}): {e}",
            ) from e
        except YAMLError as e:
            raise InvalidAPIParameters(f"Invalid YAML syntax: {e}") from e


class EndpointStatistics:
    @classmethod
    async def batch_load_by_endpoint_impl(
        cls,
        valkey_stat_client: ValkeyStatClient,
        endpoint_ids: Sequence[UUID],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        endpoint_id_strs = [str(endpoint_id) for endpoint_id in endpoint_ids]
        return await valkey_stat_client.get_inference_app_statistics_batch(endpoint_id_strs)

    @classmethod
    async def batch_load_by_endpoint(
        cls,
        ctx: "GraphQueryContext",
        endpoint_ids: Sequence[UUID],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        return await cls.batch_load_by_endpoint_impl(ctx.valkey_stat, endpoint_ids)

    @classmethod
    async def batch_load_by_replica(
        cls,
        ctx: GraphQueryContext,
        endpoint_replica_ids: Sequence[tuple[UUID, UUID]],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        endpoint_replica_pairs = [
            (str(endpoint_id), str(replica_id)) for endpoint_id, replica_id in endpoint_replica_ids
        ]
        return await ctx.valkey_stat.get_inference_replica_statistics_batch(endpoint_replica_pairs)



from __future__ import annotations

import logging
import traceback
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Iterable, Optional, Self, Tuple, cast

import aiohttp_cors
import attrs
import graphene
import strawberry
import trafaret as t
from aiohttp import web
from graphene.validation import depth_limit_validator
from graphql import ValidationRule, parse, validate
from graphql.error import GraphQLError  # pants: no-infer-dep
from graphql.execution import ExecutionResult  # pants: no-infer-dep
from pydantic import ConfigDict, Field

from ai.backend.common import validators as tx
from ai.backend.common.api_handlers import APIResponse, BodyParam, MiddlewareParam, api_handler
from ai.backend.common.dto.manager.request import GraphQLReq
from ai.backend.common.dto.manager.response import GraphQLResponse
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.dto.context import ProcessorsCtx

from ..api.gql.schema import schema as strawberry_schema
from ..errors.api import GraphQLError as BackendGQLError
from ..models.base import DataLoaderManager
from ..models.gql import (
    GQLExceptionMiddleware,
    GQLMetricMiddleware,
    GQLMutationPrivilegeCheckMiddleware,
    GraphQueryContext,
    Mutation,
    Query,
)
from .auth import auth_required, auth_required_for_method
from .manager import GQLMutationUnfrozenRequiredMiddleware
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params

if TYPE_CHECKING:
    from graphql import FieldNode

    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GQLLoggingMiddleware:
    def resolve(self, next, root, info: graphene.ResolveInfo, **args) -> Any:
        if info.path.prev is None:  # indicates the root query
            graph_ctx = info.context
            log.info(
                "ADMIN.GQL (ak:{}, {}:{}, op:{})",
                graph_ctx.access_key,
                info.operation.operation,
                info.field_name,
                info.operation.name,
            )
        return next(root, info, **args)


class CustomIntrospectionRule(ValidationRule):
    def enter_field(self, node: FieldNode, *_args):
        field_name = node.name.value
        if field_name.startswith("__"):
            # Allow __typename field for GraphQL Federation, @connection directive
            if field_name == "__typename":
                return
            self.report_error(
                GraphQLError(f"Cannot query '{field_name}': introspection is disabled.", node)
            )


async def _handle_gql_common(request: web.Request, params: Any) -> ExecutionResult:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["admin.context"]
    manager_status = await root_ctx.config_provider.legacy_etcd_config_loader.get_manager_status()
    known_slot_types = await root_ctx.config_provider.legacy_etcd_config_loader.get_resource_slots()
    rules = []
    if not root_ctx.config_provider.config.api.allow_graphql_schema_introspection:
        rules.append(CustomIntrospectionRule)
    max_depth = cast(int | None, root_ctx.config_provider.config.api.max_gql_query_depth)
    if max_depth is not None:
        rules.append(depth_limit_validator(max_depth=max_depth))
    if rules:
        validate_errors = validate(
            schema=app_ctx.gql_schema.graphql_schema,
            document_ast=parse(params["query"]),
            rules=rules,
        )
        if validate_errors:
            return ExecutionResult(None, errors=validate_errors)
    gql_ctx = GraphQueryContext(
        schema=app_ctx.gql_schema,
        dataloader_manager=DataLoaderManager(),
        config_provider=root_ctx.config_provider,
        etcd=root_ctx.etcd,
        user=request["user"],
        access_key=request["keypair"]["access_key"],
        db=root_ctx.db,
        valkey_stat=root_ctx.valkey_stat,
        valkey_image=root_ctx.valkey_image,
        valkey_live=root_ctx.valkey_live,
        valkey_schedule=root_ctx.valkey_schedule,
        network_plugin_ctx=root_ctx.network_plugin_ctx,
        manager_status=manager_status,
        known_slot_types=known_slot_types,
        background_task_manager=root_ctx.background_task_manager,
        services_ctx=root_ctx.services_ctx,
        storage_manager=root_ctx.storage_manager,
        registry=root_ctx.registry,
        idle_checker_host=root_ctx.idle_checker_host,
        metric_observer=root_ctx.metrics.gql,
        processors=root_ctx.processors,
        scheduler_repository=root_ctx.repositories.scheduler.repository,
        user_repository=root_ctx.repositories.user.repository,
        agent_repository=root_ctx.repositories.agent.repository,
    )
    result = await app_ctx.gql_schema.execute_async(
        params["query"],
        None,  # root
        variable_values=params["variables"],
        operation_name=params["operation_name"],
        context_value=gql_ctx,
        middleware=[
            GQLMutationPrivilegeCheckMiddleware(),
            GQLMutationUnfrozenRequiredMiddleware(),
            GQLMetricMiddleware(),
            GQLExceptionMiddleware(),
            GQLLoggingMiddleware(),
        ],
    )

    if result.errors:
        for e in result.errors:
            if isinstance(e, GraphQLError):
                errmsg = e.formatted
            else:
                errmsg = {"message": str(e)}
            log.error("ADMIN.GQL Exception: {}", errmsg)
            log.debug("{}", "".join(traceback.format_exception(e)))
    return result


@auth_required
@check_api_params(
    t.Dict({
        t.Key("query"): t.String,
        t.Key("variables", default=None): t.Null | t.Mapping(t.String, t.Any),
        tx.AliasedKey(["operation_name", "operationName"], default=None): t.Null | t.String,
    })
)
async def handle_gql_graphene(request: web.Request, params: Any) -> web.Response:
    result = await _handle_gql_common(request, params)
    return web.json_response(result.formatted, status=HTTPStatus.OK)


class GQLInspectionConfigCtx(MiddlewareParam):
    gql_v2_schema: strawberry.Schema = Field(
        ..., description="Strawberry GraphQL schema for v2 API."
    )
    allow_graphql_schema_introspection: bool
    max_gql_query_depth: Optional[int]

    # Allow strawberry.Schema to be used as a type
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    async def from_request(cls, request: web.Request) -> Self:
        root_ctx: RootContext = request.app["_root.context"]
        app_ctx: PrivateContext = request.app["admin.context"]

        return cls(
            gql_v2_schema=app_ctx.gql_v2_schema,
            allow_graphql_schema_introspection=root_ctx.config_provider.config.api.allow_graphql_schema_introspection,
            max_gql_query_depth=root_ctx.config_provider.config.api.max_gql_query_depth,
        )


class ConfigProviderCtx(MiddlewareParam):
    config_provider: ManagerConfigProvider

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    async def from_request(cls, request: web.Request) -> Self:
        root_ctx: RootContext = request.app["_root.context"]

        return cls(
            config_provider=root_ctx.config_provider,
        )


class GQLAPIHandler:
    @auth_required_for_method
    @api_handler
    async def handle_gql_strawberry(
        self,
        body: BodyParam[GraphQLReq],
        gql_config_ctx: GQLInspectionConfigCtx,
        config_provider_ctx: ConfigProviderCtx,
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        rules = []

        if not gql_config_ctx.allow_graphql_schema_introspection:
            rules.append(CustomIntrospectionRule)
        max_depth = cast(int | None, gql_config_ctx.max_gql_query_depth)
        if max_depth is not None:
            rules.append(depth_limit_validator(max_depth=max_depth))

        if rules:
            validate_errors = validate(
                # TODO: Instead of accessing private field, use another approach
                schema=gql_config_ctx.gql_v2_schema._schema,
                document_ast=parse(body.parsed.query),
                rules=rules,
            )
            if validate_errors:
                validation_result = ExecutionResult(None, errors=validate_errors)
                return APIResponse.build(
                    status_code=HTTPStatus.BAD_REQUEST,
                    response_model=GraphQLResponse(
                        data=None, errors=validation_result.formatted.get("errors", [])
                    ),
                )

        strawberry_ctx = StrawberryGQLContext(
            processors=processors_ctx.processors,
            config_provider=config_provider_ctx.config_provider,
        )

        query, variables, operation_name = (
            body.parsed.query,
            body.parsed.variables,
            body.parsed.operation_name,
        )

        result = await gql_config_ctx.gql_v2_schema.execute(
            query,
            root_value=None,
            variable_values=variables,
            operation_name=operation_name,
            context_value=strawberry_ctx,
        )

        errors = []
        if result.errors:
            for err in result.errors:
                log.error("ADMIN.GQL-V2 Exception: {}", err.formatted)
                errors.append(err.formatted)

        response_data = GraphQLResponse(
            data=result.data, errors=errors, extensions=result.extensions
        )
        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response_data,
        )


@auth_required
@check_api_params(
    t.Dict({
        t.Key("query"): t.String,
        t.Key("variables", default=None): t.Null | t.Mapping(t.String, t.Any),
        tx.AliasedKey(["operation_name", "operationName"], default=None): t.Null | t.String,
    })
)
async def handle_gql_legacy(request: web.Request, params: Any) -> web.Response:
    # FIXME: remove in v21.09
    result = await _handle_gql_common(request, params)
    if result.errors:
        errors = []
        for e in result.errors:
            if isinstance(e, GraphQLError):
                errmsg = e.formatted
                errors.append(errmsg)
            else:
                errmsg = {"message": str(e)}
                errors.append(errmsg)
            log.error("ADMIN.GQL Exception: {}", errmsg)
        raise BackendGQLError(extra_data=errors)
    return web.json_response(result.data, status=HTTPStatus.OK)


@attrs.define(auto_attribs=True, slots=True, init=False)
class PrivateContext:
    gql_schema: graphene.Schema
    gql_v2_schema: strawberry.Schema


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["admin.context"]
    app_ctx.gql_schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
        auto_camelcase=False,
    )
    app_ctx.gql_v2_schema = strawberry_schema
    root_ctx: RootContext = app["_root.context"]
    if root_ctx.config_provider.config.api.allow_graphql_schema_introspection:
        log.warning(
            "GraphQL schema introspection is enabled. "
            "It is strongly advised to disable this in production setups."
        )


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["admin.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", r"/graphql", handle_gql_legacy))
    cors.add(app.router.add_route("POST", r"/gql", handle_gql_graphene))

    gql_api_handler = GQLAPIHandler()
    cors.add(
        app.router.add_route("POST", r"/gql/strawberry", gql_api_handler.handle_gql_strawberry)
    )
    return app, []



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.model_serving.types import CompactServiceInfo
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class ListModelServiceAction(ModelServiceAction):
    session_owener_id: uuid.UUID
    name: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list"


@dataclass
class ListModelServiceActionResult(BaseActionResult):
    data: list[CompactServiceInfo]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.auth.types import SSHKeypair
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class UploadSSHKeypairAction(AuthAction):
    user_id: uuid.UUID
    public_key: str
    private_key: str
    access_key: str

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "upload_ssh_keypair"


@dataclass
class UploadSSHKeypairActionResult(BaseActionResult):
    ssh_keypair: SSHKeypair

    @override
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import ArtifactDataWithRevisions
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class UpsertArtifactsAction(ArtifactAction):
    data: list[ArtifactDataWithRevisions]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "upsert_multi"


@dataclass
class UpsertArtifactsActionResult(BaseActionResult):
    result: list[ArtifactDataWithRevisions]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

import logging
from pprint import pformat, pprint
from typing import Any, Optional

import click
import sqlalchemy as sa
from tabulate import tabulate

from ai.backend.common.arch import CURRENT_ARCH
from ai.backend.common.docker import validate_image_labels
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import ImageAlias
from ai.backend.logging import BraceStyleAdapter

from ..data.image.types import ImageStatus
from ..models.image import ImageAliasRow, ImageIdentifier, ImageRow
from ..models.image import rescan_images as rescan_images_func
from ..models.utils import connect_database
from .context import CLIContext, redis_ctx

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


async def list_images(cli_ctx: CLIContext, short, installed_only):
    # Connect to postgreSQL DB
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_readonly_session() as session,
        redis_ctx(cli_ctx) as redis_conn_set,
    ):
        displayed_items: list[tuple[Any, ...]] = []
        try:
            # Idea: Add `--include-deleted` option to include deleted images?
            items = await ImageRow.list(session)
            # NOTE: installed/installed_agents fields are no longer provided in CLI,
            #       until we finish the epic refactoring of image metadata db.
            if installed_only:
                image_canonicals = [item.name for item in items]
                installed_counts = await redis_conn_set.image.get_agent_counts_for_images(
                    image_canonicals
                )
                installed_items = []

                for item, installed_count in zip(items, installed_counts):
                    if installed_count > 0:
                        installed_items.append(item)

                installed_canonicals = [item.name for item in installed_items]
                agents_per_installed_items = await redis_conn_set.image.get_agents_for_images(
                    installed_canonicals
                )

                for item, installed_agents in zip(installed_items, agents_per_installed_items):
                    formatted_installed_agents = " ".join(installed_agents)
                    if short:
                        displayed_items.append((
                            item.image_ref.canonical,
                            item.config_digest,
                            formatted_installed_agents,
                        ))
                    else:
                        print(f"{pformat(item)} @ {formatted_installed_agents}")
            else:
                for item in items:
                    if short:
                        displayed_items.append((item.image_ref.canonical, item.config_digest))
                    else:
                        pprint(item)
            if short:
                print(tabulate(displayed_items, tablefmt="plain"))
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def inspect_image(cli_ctx: CLIContext, canonical_or_alias, architecture):
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            pprint(await image_row.inspect())
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def forget_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.bootstrap_config.db) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            await image_row.mark_as_deleted(session)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def purge_image(
    cli_ctx: CLIContext, canonical_or_alias: str, architecture: str, remove_from_registry: bool
):
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
                filter_by_statuses=None,
            )
            await session.delete(image_row)

            if remove_from_registry:
                await image_row.untag_image_from_registry(db=db, session=session)

        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def set_image_resource_limit(
    cli_ctx: CLIContext,
    canonical_or_alias,
    slot_type,
    range_value,
    architecture,
):
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            await image_row.set_resource_limit(slot_type, range_value)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def rescan_images(
    cli_ctx: CLIContext, registry_or_image: str, project: Optional[str] = None
) -> None:
    if not registry_or_image:
        raise click.BadArgumentUsage("Please specify a valid registry or full image name.")
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
    ):
        try:
            result = await rescan_images_func(db, registry_or_image, project)
            for error in result.errors:
                log.error(f"Failed to scan registries: {error}")
        except Exception as e:
            log.exception(f"Unknown error occurred. Error: {e}")


async def alias(cli_ctx: CLIContext, alias, target, architecture):
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(target, architecture),
                ],
            )
            await ImageAliasRow.create(session, alias, image_row)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def dealias(cli_ctx: CLIContext, alias):
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_session() as session,
    ):
        alias_row = await session.scalar(
            sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
        )
        if alias_row is None:
            log.exception("Alias not found.")
            return
        await session.delete(alias_row)


async def validate_image_alias(cli_ctx: CLIContext, alias: str) -> None:
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            image_row = await ImageRow.from_alias(session, alias)
            for key, value in validate_image_labels(image_row.labels).items():
                print(f"{key:<40}: ", end="")
                if isinstance(value, list):
                    value = f"[{', '.join(value)}]"
                print(value)

        except UnknownImageReference:
            log.error(f"No images were found with alias: {alias}")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


def _resolve_architecture(current: bool, architecture: Optional[str]) -> str:
    if architecture is not None:
        return architecture
    if current:
        return CURRENT_ARCH

    raise ValueError("Unreachable code!")


async def validate_image_canonical(
    cli_ctx: CLIContext, canonical: str, current: bool, architecture: Optional[str] = None
) -> None:
    bootstrap_config = await cli_ctx.get_bootstrap_config()
    async with (
        connect_database(bootstrap_config.db) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            if current or architecture is not None:
                resolved_arch = _resolve_architecture(current, architecture)
                image_row = await ImageRow.resolve(
                    session, [ImageIdentifier(canonical, resolved_arch)]
                )

                print(f"{'architecture':<40}: {resolved_arch}")
                for key, value in validate_image_labels(image_row.labels).items():
                    print(f"{key:<40}: ", end="")
                    if isinstance(value, list):
                        value = f"{', '.join(value)}"
                    print(value)
            else:
                rows = await session.scalars(
                    sa.select(ImageRow).where(
                        sa.and_(ImageRow.name == canonical, ImageRow.status == ImageStatus.ALIVE)
                    )
                )
                image_rows = rows.fetchall()
                if not image_rows:
                    raise UnknownImageReference(f"{canonical}")
                for i, image_row in enumerate(image_rows):
                    if i > 0:
                        print("-" * 50)
                    print(f"{'architecture':<40}: {image_row.architecture}")
                    for key, value in validate_image_labels(image_row.labels).items():
                        print(f"{key:<40}: ", end="")
                        if isinstance(value, list):
                            value = f"{', '.join(value)}"
                        print(value)

        except UnknownImageReference as e:
            log.error(f"{e}")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")



import logging
from collections.abc import Mapping
from typing import Any

from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.types import LegacyResourceSlotState as ResourceSlotState
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.repositories.resource_preset import ResourcePresetRepository
from ai.backend.manager.services.resource_preset.actions.check_presets import (
    CheckResourcePresetsAction,
    CheckResourcePresetsActionResult,
)
from ai.backend.manager.services.resource_preset.actions.create_preset import (
    CreateResourcePresetAction,
    CreateResourcePresetActionResult,
)
from ai.backend.manager.services.resource_preset.actions.delete_preset import (
    DeleteResourcePresetAction,
    DeleteResourcePresetActionResult,
)
from ai.backend.manager.services.resource_preset.actions.list_presets import (
    ListResourcePresetsAction,
    ListResourcePresetsResult,
)
from ai.backend.manager.services.resource_preset.actions.modify_preset import (
    ModifyResourcePresetAction,
    ModifyResourcePresetActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ResourcePresetService:
    _resource_preset_repository: ResourcePresetRepository

    def __init__(
        self,
        resource_preset_repository: ResourcePresetRepository,
    ) -> None:
        self._resource_preset_repository = resource_preset_repository

    async def create_preset(
        self, action: CreateResourcePresetAction
    ) -> CreateResourcePresetActionResult:
        creator = action.creator

        if not creator.resource_slots.has_intrinsic_slots():
            raise InvalidAPIParameters("ResourceSlot must have all intrinsic resource slots.")

        preset_data = await self._resource_preset_repository.create_preset_validated(creator)
        return CreateResourcePresetActionResult(resource_preset=preset_data)

    async def modify_preset(
        self, action: ModifyResourcePresetAction
    ) -> ModifyResourcePresetActionResult:
        name = action.name
        preset_id = action.id
        modifier = action.modifier

        if preset_id is None and name is None:
            raise InvalidAPIParameters("One of (`id` or `name`) parameter should not be null")

        if resource_slots := modifier.resource_slots.optional_value():
            if not resource_slots.has_intrinsic_slots():
                raise InvalidAPIParameters("ResourceSlot must have all intrinsic resource slots.")

        preset_data = await self._resource_preset_repository.modify_preset_validated(
            preset_id, name, modifier
        )

        return ModifyResourcePresetActionResult(resource_preset=preset_data)

    async def delete_preset(
        self, action: DeleteResourcePresetAction
    ) -> DeleteResourcePresetActionResult:
        name = action.name
        preset_id = action.id

        if preset_id is None and name is None:
            raise InvalidAPIParameters("One of (`id` or `name`) parameter should not be null")

        preset_data = await self._resource_preset_repository.delete_preset_validated(
            preset_id, name
        )

        return DeleteResourcePresetActionResult(resource_preset=preset_data)

    async def list_presets(self, action: ListResourcePresetsAction) -> ListResourcePresetsResult:
        preset_data_list = await self._resource_preset_repository.list_presets(action.scaling_group)

        presets = []
        for preset_data in preset_data_list:
            preset_slots = preset_data.resource_slots.normalize_slots(ignore_unknown=True)
            presets.append({
                "id": str(preset_data.id),
                "name": preset_data.name,
                "shared_memory": str(preset_data.shared_memory)
                if preset_data.shared_memory
                else None,
                "resource_slots": preset_slots.to_json(),
            })

        return ListResourcePresetsResult(presets=presets)

    async def check_presets(
        self, action: CheckResourcePresetsAction
    ) -> CheckResourcePresetsActionResult:
        result = await self._resource_preset_repository.check_presets(
            access_key=action.access_key,
            user_id=action.user_id,
            group_name=action.group,
            domain_name=action.domain_name,
            resource_policy=action.resource_policy,
            scaling_group=action.scaling_group,
        )

        # Convert repository result to action result
        # Process presets to JSON format
        presets: list[Mapping[str, Any]] = []
        for preset_data in result.presets:
            preset_slots = preset_data.preset.resource_slots.normalize_slots(ignore_unknown=True)
            presets.append({
                "id": str(preset_data.preset.id),
                "name": preset_data.preset.name,
                "resource_slots": preset_slots.to_json(),
                "shared_memory": (
                    str(preset_data.preset.shared_memory)
                    if preset_data.preset.shared_memory is not None
                    else None
                ),
                "allocatable": preset_data.allocatable,
            })

        # Convert per scaling group data to appropriate format
        per_sgroup_dict = {}
        for sgname, sg_data in result.scaling_groups.items():
            per_sgroup_dict[sgname] = {
                ResourceSlotState.OCCUPIED: sg_data.using,
                ResourceSlotState.AVAILABLE: sg_data.remaining,
            }

        return CheckResourcePresetsActionResult(
            presets=presets,
            keypair_limits=result.keypair_limits,
            keypair_using=result.keypair_using,
            keypair_remaining=result.keypair_remaining,
            group_limits=result.group_limits,
            group_using=result.group_using,
            group_remaining=result.group_remaining,
            scaling_group_remaining=result.scaling_group_remaining,
            scaling_groups=per_sgroup_dict,
        )



from __future__ import annotations

import asyncio
import enum
import logging
from collections.abc import Iterable, Mapping, Sequence
from contextlib import asynccontextmanager as actxmgr
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    List,
    Optional,
    Self,
    TypeAlias,
    Union,
    cast,
    override,
)
from uuid import UUID

import aiotools
import redis.exceptions
import sqlalchemy as sa
from dateutil.tz import tzutc
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import (
    contains_eager,
    foreign,
    joinedload,
    load_only,
    noload,
    relationship,
    selectinload,
)

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.defs.session import SESSION_PRIORITY_DEFAULT
from ai.backend.common.events.dispatcher import (
    EventProducer,
)
from ai.backend.common.events.event_types.schedule.anycast import (
    DoStartSessionEvent,
)
from ai.backend.common.events.event_types.session.anycast import (
    DoUpdateSessionStatusEvent,
    SessionStartedAnycastEvent,
    SessionTerminatedAnycastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SessionStartedBroadcastEvent,
    SessionTerminatedBroadcastEvent,
)
from ai.backend.common.plugin.hook import HookPluginContext
from ai.backend.common.types import (
    AccessKey,
    ClusterMode,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionResult,
    SessionTypes,
    VFolderMount,
)
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.user.types import UserData

if TYPE_CHECKING:
    from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import BackendAIError
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.session.types import (
    ImageSpec,
    MountSpec,
    ResourceSpec,
    SessionData,
    SessionExecution,
    SessionIdentity,
    SessionInfo,
    SessionLifecycle,
    SessionMetadata,
    SessionMetrics,
    SessionNetwork,
    SessionStatus,
)

from ..defs import DEFAULT_ROLE
from ..errors.kernel import (
    KernelCreationFailed,
    KernelDestructionFailed,
    KernelExecutionFailed,
    KernelNotFound,
    KernelRestartFailed,
    MainKernelNotFound,
    SessionNotFound,
    TooManyKernelsFound,
    TooManySessionsMatched,
)
from ..exceptions import AgentError
from .base import (
    GUID,
    Base,
    EnumType,
    ForeignKeyIDColumn,
    ResourceSlotColumn,
    SessionIDColumn,
    StrEnumType,
    StructuredJSONObjectListColumn,
    URLColumn,
)
from .group import GroupRow
from .image import ImageRow
from .kernel import KernelRow
from .minilang.queryfilter import FieldSpecType, QueryFilterParser
from .network import NetworkRow, NetworkType
from .rbac import (
    AbstractPermissionContext,
    AbstractPermissionContextBuilder,
    DomainScope,
    ProjectScope,
    ScopeType,
    get_predefined_roles_in_scope,
)
from .rbac import (
    UserScope as UserRBACScope,
)
from .rbac.context import ClientContext
from .rbac.permission_defs import ComputeSessionPermission
from .routing import RouteStatus, RoutingRow
from .types import (
    QueryCondition,
    QueryOption,
)
from .utils import (
    ExtendedAsyncSAEngine,
    JSONCoalesceExpr,
    execute_with_retry,
    execute_with_txn_retry,
    sql_json_merge,
)

if TYPE_CHECKING:
    from ..registry import AgentRegistry

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "determine_session_status_by_kernels",
    "handle_session_exception",
    "ALLOWED_IMAGE_ROLES_FOR_SESSION_TYPE",
    "PRIVATE_SESSION_TYPES",
    "SESSION_STATUS_TRANSITION_MAP",
    "DEAD_SESSION_STATUSES",
    "AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES",
    "USER_RESOURCE_OCCUPYING_SESSION_STATUSES",
    "SessionRow",
    "SessionDependencyRow",
    "check_all_dependencies",
    "KernelLoadingStrategy",
)

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models.session"))


FOLLOWING_SESSION_STATUSES = (
    # Session statuses that need to wait all kernels belonging to the session
    SessionStatus.PREPARED,
    SessionStatus.RUNNING,
    SessionStatus.TERMINATED,
)
LEADING_SESSION_STATUSES = tuple(
    # Session statuses that declare first, do not need to wait any sibling kernel
    s
    for s in SessionStatus
    if s not in FOLLOWING_SESSION_STATUSES
)

DEAD_SESSION_STATUSES = frozenset([
    SessionStatus.CANCELLED,
    SessionStatus.TERMINATED,
])

# statuses to consider when calculating current resource usage
AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES = tuple(
    e
    for e in SessionStatus
    if e
    not in (
        SessionStatus.TERMINATED,
        SessionStatus.PENDING,
        SessionStatus.CANCELLED,
    )
)

# statuses that occupy user resources
# these statuses are used to calculate user resource usage
USER_RESOURCE_OCCUPYING_SESSION_STATUSES = tuple(
    e
    for e in SessionStatus
    if e
    not in (
        SessionStatus.TERMINATED,
        SessionStatus.PENDING,
        SessionStatus.CANCELLED,
    )
)

PRIVATE_SESSION_TYPES = (SessionTypes.SYSTEM,)

OP_EXC = {
    "create_session": KernelCreationFailed,
    "restart_session": KernelRestartFailed,
    "destroy_session": KernelDestructionFailed,
    "execute": KernelExecutionFailed,
    "shutdown_service": KernelExecutionFailed,
    "upload_file": KernelExecutionFailed,
    "download_file": KernelExecutionFailed,
    "download_single": KernelExecutionFailed,
    "list_files": KernelExecutionFailed,
    "get_logs_from_agent": KernelExecutionFailed,
    "refresh_session": KernelExecutionFailed,
    "commit_session": KernelExecutionFailed,
    "commit_session_to_file": KernelExecutionFailed,
    "trigger_batch_execution": KernelExecutionFailed,
}


KERNEL_SESSION_STATUS_MAPPING: Mapping[KernelStatus, SessionStatus] = {
    KernelStatus.PENDING: SessionStatus.PENDING,
    KernelStatus.SCHEDULED: SessionStatus.SCHEDULED,
    KernelStatus.PREPARING: SessionStatus.PREPARING,
    KernelStatus.BUILDING: SessionStatus.PREPARING,
    KernelStatus.PULLING: SessionStatus.PULLING,
    KernelStatus.PREPARED: SessionStatus.PREPARED,
    KernelStatus.CREATING: SessionStatus.CREATING,
    KernelStatus.RUNNING: SessionStatus.RUNNING,
    KernelStatus.RESTARTING: SessionStatus.RESTARTING,
    KernelStatus.RESIZING: SessionStatus.RUNNING,
    KernelStatus.SUSPENDED: SessionStatus.ERROR,
    KernelStatus.TERMINATING: SessionStatus.TERMINATING,
    KernelStatus.TERMINATED: SessionStatus.TERMINATED,
    KernelStatus.ERROR: SessionStatus.ERROR,
    KernelStatus.CANCELLED: SessionStatus.CANCELLED,
}

SESSION_KERNEL_STATUS_MAPPING: Mapping[SessionStatus, KernelStatus] = {
    SessionStatus.PENDING: KernelStatus.PENDING,
    SessionStatus.SCHEDULED: KernelStatus.SCHEDULED,
    SessionStatus.PREPARING: KernelStatus.PREPARING,
    SessionStatus.PULLING: KernelStatus.PULLING,
    SessionStatus.PREPARED: KernelStatus.PREPARED,
    SessionStatus.CREATING: KernelStatus.CREATING,
    SessionStatus.RUNNING: KernelStatus.RUNNING,
    SessionStatus.RESTARTING: KernelStatus.RESTARTING,
    SessionStatus.TERMINATING: KernelStatus.TERMINATING,
    SessionStatus.TERMINATED: KernelStatus.TERMINATED,
    SessionStatus.ERROR: KernelStatus.ERROR,
    SessionStatus.CANCELLED: KernelStatus.CANCELLED,
}

SESSION_STATUS_TRANSITION_MAP: Mapping[SessionStatus, set[SessionStatus]] = {
    SessionStatus.PENDING: {
        SessionStatus.SCHEDULED,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.SCHEDULED: {
        SessionStatus.PREPARING,
        SessionStatus.PULLING,
        SessionStatus.PREPARED,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.PREPARING: {
        SessionStatus.PULLING,
        SessionStatus.PREPARED,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.PULLING: {
        SessionStatus.PREPARED,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.PREPARED: {
        SessionStatus.PREPARING,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.CREATING: {
        SessionStatus.RUNNING,
        SessionStatus.ERROR,
        SessionStatus.CANCELLED,
    },
    SessionStatus.RUNNING: {
        SessionStatus.RESTARTING,
        SessionStatus.RUNNING_DEGRADED,
        SessionStatus.TERMINATING,
        SessionStatus.TERMINATED,
        SessionStatus.ERROR,
    },
    SessionStatus.RESTARTING: {
        SessionStatus.RUNNING,
        SessionStatus.RUNNING_DEGRADED,
        SessionStatus.TERMINATING,
        SessionStatus.TERMINATED,
        SessionStatus.ERROR,
    },
    SessionStatus.RUNNING_DEGRADED: {
        SessionStatus.RUNNING,
        SessionStatus.TERMINATING,
        SessionStatus.TERMINATED,
        SessionStatus.ERROR,
    },
    SessionStatus.TERMINATING: {SessionStatus.TERMINATED, SessionStatus.ERROR},
    SessionStatus.TERMINATED: set(),
    SessionStatus.ERROR: {SessionStatus.TERMINATING, SessionStatus.TERMINATED},
    SessionStatus.CANCELLED: set(),
}


# TODO:
def determine_session_status_by_kernels(kernels: Sequence[KernelRow]) -> SessionStatus:
    if not kernels:
        raise KernelNotFound
    candidate = KERNEL_SESSION_STATUS_MAPPING[kernels[0].status]
    if len(kernels) == 1:
        return candidate

    for k in kernels:
        match k.status:
            case KernelStatus.ERROR:
                # If any kernel status is ERROR, determines session status as ERROR
                return SessionStatus.ERROR
            case (
                KernelStatus.BUILDING
                | KernelStatus.RESTARTING
                | KernelStatus.RESIZING
                | KernelStatus.SUSPENDED
            ):
                raise RuntimeError("Status not used.")

        match candidate:
            case SessionStatus.PENDING:
                match k.status:
                    case KernelStatus.PENDING:
                        continue
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.SCHEDULED:
                match k.status:
                    case KernelStatus.SCHEDULED | KernelStatus.PREPARED:
                        continue
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case KernelStatus.PULLING:
                        candidate = SessionStatus.PULLING
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.PREPARING:
                match k.status:
                    case KernelStatus.PREPARING | KernelStatus.PREPARED:
                        continue
                    case KernelStatus.PULLING:
                        candidate = SessionStatus.PULLING
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.PULLING:
                match k.status:
                    case KernelStatus.PULLING | KernelStatus.PREPARING | KernelStatus.PREPARED:
                        continue
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.PREPARED:
                match k.status:
                    case KernelStatus.PREPARED:
                        continue
                    case KernelStatus.PREPARING:
                        candidate = SessionStatus.PREPARING
                    case KernelStatus.PULLING:
                        candidate = SessionStatus.PULLING
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.CREATING:
                match k.status:
                    case KernelStatus.CREATING | KernelStatus.RUNNING:
                        continue
                    case KernelStatus.CANCELLED:
                        candidate = SessionStatus.CANCELLED
                    case _:
                        # Set status to ERROR if any kernel is in exceptional state
                        return SessionStatus.ERROR
            case SessionStatus.CANCELLED:
                match k.status:
                    case (
                        KernelStatus.CANCELLED
                        | KernelStatus.PENDING
                        | KernelStatus.SCHEDULED
                        | KernelStatus.PREPARING
                        | KernelStatus.PULLING
                        | KernelStatus.PREPARED
                    ):
                        continue
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.RUNNING:
                match k.status:
                    case KernelStatus.RUNNING:
                        continue
                    case KernelStatus.CREATING:
                        candidate = SessionStatus.CREATING
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.TERMINATING:
                match k.status:
                    case KernelStatus.TERMINATING | KernelStatus.TERMINATED:
                        continue
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.TERMINATED:
                match k.status:
                    case KernelStatus.TERMINATED:
                        continue
                    case KernelStatus.TERMINATING:
                        candidate = SessionStatus.TERMINATING
                    case _:
                        return SessionStatus.ERROR
            case SessionStatus.RESTARTING | SessionStatus.RUNNING_DEGRADED:
                raise RuntimeError("Status not used.")
    return candidate


@actxmgr
async def handle_session_exception(
    db: ExtendedAsyncSAEngine,
    op: str,
    session_id: SessionId,
    error_callback=None,
    cancellation_callback=None,
    set_error: bool = False,
) -> AsyncIterator[None]:
    exc_class = OP_EXC[op]
    try:
        yield
    except asyncio.TimeoutError:
        if set_error:
            await SessionRow.set_session_status(
                db,
                session_id,
                SessionStatus.ERROR,
                reason=f"operation-timeout ({op})",
            )
        if error_callback:
            await error_callback()
        raise exc_class("TIMEOUT") from None
    except asyncio.CancelledError:
        if cancellation_callback:
            await cancellation_callback()
        raise
    except AgentError as e:
        if set_error:
            await SessionRow.set_session_status(
                db,
                session_id,
                SessionStatus.ERROR,
                reason=f"agent-error ({e!r})",
                status_data={
                    "error": {
                        "src": "agent",
                        "agent_id": e.agent_id,
                        "name": e.exc_name,
                        "repr": e.exc_repr,
                    },
                },
            )
        if error_callback:
            await error_callback()
        raise exc_class("FAILURE", e) from None
    except BackendAIError:
        # silently re-raise to make them handled by gateway http handlers
        raise
    except Exception as e:
        if set_error:
            await SessionRow.set_session_status(
                db,
                session_id,
                SessionStatus.ERROR,
                reason=f"other-error ({e!r})",
                status_data={
                    "error": {
                        "src": "other",
                        "name": e.__class__.__name__,
                        "repr": repr(e),
                    },
                },
            )
        if error_callback:
            await error_callback()
        raise


def _build_session_fetch_query(
    base_cond,
    access_key: AccessKey | None = None,
    *,
    allow_stale: bool = True,
    for_update: bool = False,
    do_ordering: bool = False,
    max_matches: Optional[int] = None,
    eager_loading_op: Optional[Sequence] = None,
):
    cond = base_cond
    if access_key:
        cond = cond & (SessionRow.access_key == access_key)
    if not allow_stale:
        cond = cond & (~SessionRow.status.in_(DEAD_SESSION_STATUSES))
    query = (
        sa.select(SessionRow)
        .where(cond)
        .order_by(sa.desc(SessionRow.created_at))
        .execution_options(populate_existing=True)
    )
    if max_matches is not None:
        query = query.limit(max_matches).offset(0)
    if for_update:
        query = query.with_for_update()
    if do_ordering:
        query = query.order_by(SessionRow.created_at)
    if eager_loading_op is not None:
        query = query.options(*eager_loading_op)

    return query


async def _match_sessions_by_id(
    db_session: SASession,
    session_id_or_list: SessionId | list[SessionId],
    access_key: AccessKey | None = None,
    *,
    allow_prefix: bool = False,
    allow_stale: bool = True,
    for_update: bool = False,
    max_matches: Optional[int] = None,
    eager_loading_op: Optional[Sequence] = None,
) -> List[SessionRow]:
    if isinstance(session_id_or_list, list):
        cond = SessionRow.id.in_(session_id_or_list)
    else:
        if allow_prefix:
            cond = sa.sql.expression.cast(SessionRow.id, sa.String).like(f"{session_id_or_list}%")
        else:
            cond = SessionRow.id == session_id_or_list
    query = _build_session_fetch_query(
        cond,
        access_key,
        max_matches=max_matches,
        allow_stale=allow_stale,
        for_update=for_update,
        eager_loading_op=eager_loading_op,
    )

    result = await db_session.execute(query)
    return result.scalars().all()


async def _match_sessions_by_name(
    db_session: SASession,
    session_name: str,
    access_key: AccessKey,
    *,
    allow_prefix: bool = False,
    allow_stale: bool = True,
    for_update: bool = False,
    max_matches: Optional[int] = None,
    eager_loading_op: Optional[Sequence] = None,
) -> List[SessionRow]:
    if allow_prefix:
        cond = sa.sql.expression.cast(SessionRow.name, sa.String).like(f"{session_name}%")
    else:
        cond = SessionRow.name == session_name
    query = _build_session_fetch_query(
        cond,
        access_key,
        max_matches=max_matches,
        allow_stale=allow_stale,
        for_update=for_update,
        eager_loading_op=eager_loading_op,
    )
    result = await db_session.execute(query)
    return result.scalars().all()


COMPUTE_CONCURRENCY_USED_KEY_PREFIX = "keypair.concurrency_used."
SYSTEM_CONCURRENCY_USED_KEY_PREFIX = "keypair.sftp_concurrency_used."


@dataclass
class ConcurrencyUsed:
    access_key: AccessKey
    compute_session_ids: set[SessionId] = field(default_factory=set)
    system_session_ids: set[SessionId] = field(default_factory=set)

    @property
    def compute_concurrency_used_key(self) -> str:
        return f"{COMPUTE_CONCURRENCY_USED_KEY_PREFIX}{self.access_key}"

    @property
    def system_concurrency_used_key(self) -> str:
        return f"{SYSTEM_CONCURRENCY_USED_KEY_PREFIX}{self.access_key}"

    def to_cnt_map(self) -> Mapping[str, int]:
        return {
            self.compute_concurrency_used_key: len(self.compute_session_ids),
            self.system_concurrency_used_key: len(self.system_session_ids),
        }


class SessionOp(enum.StrEnum):
    CREATE = "create_session"
    DESTROY = "destroy_session"
    RESTART = "restart_session"
    EXECUTE = "execute"
    REFRESH = "refresh_session"
    SHUTDOWN_SERVICE = "shutdown_service"
    UPLOAD_FILE = "upload_file"
    DOWNLOAD_FILE = "download_file"
    LIST_FILE = "list_files"
    GET_AGENT_LOGS = "get_logs_from_agent"


class KernelLoadingStrategy(enum.StrEnum):
    ALL_KERNELS = "all"
    MAIN_KERNEL_ONLY = "main"
    NONE = "none"


ALLOWED_IMAGE_ROLES_FOR_SESSION_TYPE: Mapping[SessionTypes, tuple[str, ...]] = {
    SessionTypes.BATCH: ("COMPUTE",),
    SessionTypes.INTERACTIVE: ("COMPUTE",),
    SessionTypes.INFERENCE: ("INFERENCE",),
    SessionTypes.SYSTEM: ("SYSTEM",),
}


# Defined for avoiding circular import
def _get_keypair_row_join_condition():
    from ai.backend.manager.models.keypair import KeyPairRow

    return KeyPairRow.access_key == foreign(SessionRow.access_key)


def _get_user_row_join_condition():
    from ai.backend.manager.models.user import UserRow

    return UserRow.uuid == foreign(SessionRow.user_uuid)


class SessionRow(Base):
    __tablename__ = "sessions"
    id = SessionIDColumn()
    creation_id = sa.Column("creation_id", sa.String(length=32), unique=False, index=False)
    name = sa.Column("name", sa.String(length=64), unique=False, index=True)
    session_type = sa.Column(
        "session_type",
        StrEnumType(SessionTypes, use_name=True),
        index=True,
        nullable=False,  # previously sess_type
        default=SessionTypes.INTERACTIVE,
        server_default=SessionTypes.INTERACTIVE.name,
    )
    priority = sa.Column(
        "priority",
        sa.Integer(),
        nullable=False,
        default=SESSION_PRIORITY_DEFAULT,
        index=True,
    )

    cluster_mode = sa.Column(
        "cluster_mode",
        sa.String(length=16),
        nullable=False,
        default=ClusterMode.SINGLE_NODE,
        server_default=ClusterMode.SINGLE_NODE.name,
    )
    cluster_size = sa.Column("cluster_size", sa.Integer, nullable=False, default=1)
    agent_ids = sa.Column("agent_ids", sa.ARRAY(sa.String), nullable=True)
    designated_agent_ids = sa.Column("designated_agent_ids", sa.ARRAY(sa.String), nullable=True)
    kernels = relationship("KernelRow", back_populates="session")

    # Resource ownership
    scaling_group_name = sa.Column(
        "scaling_group_name", sa.ForeignKey("scaling_groups.name"), index=True, nullable=True
    )
    scaling_group = relationship("ScalingGroupRow", back_populates="sessions")
    target_sgroup_names = sa.Column(
        "target_sgroup_names",
        sa.ARRAY(sa.String(length=64)),
        default="{}",
        server_default="{}",
        nullable=True,
    )
    domain_name = sa.Column(
        "domain_name", sa.String(length=64), sa.ForeignKey("domains.name"), nullable=False
    )
    domain = relationship("DomainRow", back_populates="sessions")
    group_id = ForeignKeyIDColumn("group_id", "groups.id", nullable=False)
    group = relationship("GroupRow", back_populates="sessions")
    user_uuid = sa.Column(
        "user_uuid", GUID, server_default=sa.text("uuid_generate_v4()"), nullable=False
    )
    user = relationship(
        "UserRow",
        primaryjoin=_get_user_row_join_condition,
        back_populates="sessions",
        foreign_keys=[user_uuid],
    )

    access_key = sa.Column("access_key", sa.String(length=20))
    access_key_row = relationship(
        "KeyPairRow",
        primaryjoin=_get_keypair_row_join_condition,
        back_populates="sessions",
        foreign_keys=[access_key],
    )

    # `image` column is identical to kernels `image` column.
    images = sa.Column("images", sa.ARRAY(sa.String), nullable=True)
    tag = sa.Column("tag", sa.String(length=64), nullable=True)

    # Resource occupation
    # occupied_slots = sa.Column('occupied_slots', ResourceSlotColumn(), nullable=False)
    occupying_slots = sa.Column("occupying_slots", ResourceSlotColumn(), nullable=False)
    requested_slots = sa.Column("requested_slots", ResourceSlotColumn(), nullable=False)
    vfolder_mounts = sa.Column(
        "vfolder_mounts", StructuredJSONObjectListColumn(VFolderMount), nullable=True
    )
    environ = sa.Column("environ", pgsql.JSONB(), nullable=True, default={})
    bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
    use_host_network = sa.Column("use_host_network", sa.Boolean(), default=False, nullable=False)

    # Lifecycle
    # Deprecated: Not used anymore
    timeout = sa.Column("timeout", sa.BigInteger(), nullable=True)
    batch_timeout = sa.Column(
        "batch_timeout", sa.BigInteger(), nullable=True
    )  # Used to set timeout of batch sessions
    created_at = sa.Column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), index=True
    )
    terminated_at = sa.Column(
        "terminated_at", sa.DateTime(timezone=True), nullable=True, default=sa.null(), index=True
    )
    starts_at = sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True, default=sa.null())
    status = sa.Column(
        "status",
        StrEnumType(SessionStatus),
        default=SessionStatus.PENDING,
        server_default=SessionStatus.PENDING.name,
        nullable=False,
        index=True,
    )
    status_info = sa.Column("status_info", sa.Unicode(), nullable=True, default=sa.null())

    status_data = sa.Column("status_data", pgsql.JSONB(), nullable=True, default=sa.null())
    # status_data contains a JSON object that contains detailed data for the last status change.
    # During scheduling (as PENDING + ("no-available-instances" | "predicate-checks-failed")):
    # {
    #   "scheduler": {
    #     // shceudler attempt information
    #     // NOTE: the whole field may be NULL before the first attempt!
    #     "retries": 5,
    #         // the number of scheudling attempts (used to avoid HoL blocking as well)
    #     "last_try": "2021-05-01T12:34:56.123456+09:00",
    #         // an ISO 8601 formatted timestamp of the last attempt
    #     "failed_predicates": [
    #       { "name": "concurrency", "msg": "You cannot run more than 30 concurrent sessions." },
    #           // see the manager.scheduler.predicates module for possible messages
    #       ...
    #     ],
    #     "passed_predicates": [ {"name": "reserved_time"}, ... ],  // names only
    #   }
    # }
    #
    # While running: the field is NULL.
    #
    # After termination:
    # {
    #   "kernel": {
    #     // termination info for the individual kernel
    #     "exit_code": 123,
    #         // maybe null during termination
    #   },
    #   "session": {
    #     // termination info for the session
    #     "status": "terminating" | "terminated"
    #         // "terminated" means all kernels that belong to the same session has terminated.
    #         // used to prevent duplication of SessionTerminatedEvent
    #   }
    # }
    status_history = sa.Column("status_history", pgsql.JSONB(), nullable=True, default=sa.null())
    callback_url = sa.Column("callback_url", URLColumn, nullable=True, default=sa.null())

    startup_command = sa.Column("startup_command", sa.Text, nullable=True)
    result = sa.Column(
        "result",
        EnumType(SessionResult),
        default=SessionResult.UNDEFINED,
        server_default=SessionResult.UNDEFINED.name,
        nullable=False,
        index=True,
    )

    # Resource metrics measured upon termination
    num_queries = sa.Column("num_queries", sa.BigInteger(), default=0)
    last_stat = sa.Column("last_stat", pgsql.JSONB(), nullable=True, default=sa.null())

    network_type = sa.Column("network_type", StrEnumType(NetworkType), nullable=True)
    """Setting this column to null means this session does not utilize inter-container networking feature"""
    network_id = sa.Column("network_id", sa.String(length=128), nullable=True)
    """
    Depending on the network_type, this column may contain a network ID or other information.
    Use `get_network_ref()` method to reveal actual network ref (generated by network plugin).
    """

    routing = relationship("RoutingRow", back_populates="session_row")

    __table_args__ = (
        # indexing
        sa.Index(
            "ix_sessions_updated_order",
            sa.func.greatest(
                "created_at",
                "terminated_at",
            ),
            unique=False,
        ),
        sa.Index("ix_sessions_vfolder_mounts", "vfolder_mounts", postgresql_using="gin"),
        sa.Index("ix_session_status_with_priority", "status", "priority"),
        # Unique index for session names excluding terminal statuses
        sa.Index(
            "ix_sessions_unique_name_nonterminal",
            "name",
            unique=True,
            postgresql_where=sa.text("status NOT IN ('ERROR', 'TERMINATED', 'CANCELLED')"),
        ),
    )

    @classmethod
    def kernel_load_option(cls, already_joined: bool = False) -> sa.orm.Load:
        return selectinload(cls.kernels) if not already_joined else contains_eager(cls.kernels)

    @classmethod
    def user_load_option(cls, already_joined: bool = False) -> sa.orm.Load:
        return joinedload(cls.user) if not already_joined else contains_eager(cls.user)

    @classmethod
    def project_load_option(cls, already_joined: bool = False) -> sa.orm.Load:
        return joinedload(cls.group) if not already_joined else contains_eager(cls.group)

    @classmethod
    def from_dataclass(cls, session_data: SessionData) -> SessionRow:
        vfolder_mounts = []
        if session_data.vfolder_mounts:
            vfolder_mounts = [
                VFolderMount.from_dataclass(mount) for mount in session_data.vfolder_mounts
            ]

        instance = cls(
            name=session_data.name,
            session_type=session_data.session_type,
            priority=session_data.priority,
            cluster_mode=session_data.cluster_mode,
            cluster_size=session_data.cluster_size,
            agent_ids=session_data.agent_ids,
            scaling_group_name=session_data.scaling_group_name,
            target_sgroup_names=session_data.target_sgroup_names,
            domain_name=session_data.domain_name,
            group_id=session_data.group_id,
            user_uuid=session_data.user_uuid,
            access_key=session_data.access_key,
            images=session_data.images,
            tag=session_data.tag,
            occupying_slots=session_data.occupying_slots,
            requested_slots=session_data.requested_slots,
            vfolder_mounts=vfolder_mounts,
            environ=session_data.environ,
            bootstrap_script=session_data.bootstrap_script,
            use_host_network=session_data.use_host_network,
            timeout=session_data.timeout,
            batch_timeout=session_data.batch_timeout,
            status_history={},
            status=session_data.status,
            status_info=session_data.status_info,
            status_data=session_data.status_data,
            callback_url=session_data.callback_url,
            startup_command=session_data.startup_command,
            result=session_data.result,
            num_queries=session_data.num_queries,
            last_stat=session_data.last_stat,
            network_type=session_data.network_type,
            network_id=session_data.network_id,
            created_at=session_data.created_at,
            terminated_at=session_data.terminated_at,
            starts_at=session_data.starts_at,
        )
        instance.id = session_data.id
        return instance

    def to_dataclass(self, owner: Optional[UserData] = None) -> SessionData:
        return SessionData(
            id=self.id,
            creation_id=self.creation_id,
            name=self.name,
            session_type=self.session_type,
            priority=self.priority,
            cluster_mode=self.cluster_mode,
            cluster_size=self.cluster_size,
            agent_ids=self.agent_ids,
            scaling_group_name=self.scaling_group_name,
            target_sgroup_names=self.target_sgroup_names,
            domain_name=self.domain_name,
            group_id=self.group_id,
            user_uuid=self.user_uuid,
            access_key=self.access_key,
            images=self.images,
            tag=self.tag,
            occupying_slots=self.occupying_slots,
            requested_slots=self.requested_slots,
            vfolder_mounts=[mount.to_dataclass() for mount in self.vfolder_mounts],
            environ=self.environ,
            bootstrap_script=self.bootstrap_script,
            use_host_network=self.use_host_network,
            timeout=self.timeout,
            batch_timeout=self.batch_timeout,
            created_at=self.created_at,
            terminated_at=self.terminated_at,
            starts_at=self.starts_at,
            status=self.status,
            status_info=self.status_info,
            status_data=self.status_data,
            status_history=self.status_history,
            callback_url=self.callback_url,
            startup_command=self.startup_command,
            result=self.result,
            num_queries=self.num_queries,
            last_stat=self.last_stat,
            network_type=self.network_type,
            network_id=self.network_id,
            service_ports=self.main_kernel.service_ports,
            owner=owner if owner is not None else None,
        )

    @classmethod
    def from_session_info(cls, info: SessionInfo) -> Self:
        return cls(
            id=info.identity.id,
            creation_id=info.identity.creation_id,
            name=info.identity.name,
            session_type=info.identity.session_type,
            priority=info.identity.priority,
            cluster_mode=info.resource.cluster_mode,
            cluster_size=info.resource.cluster_size,
            agent_ids=info.resource.agent_ids,
            scaling_group_name=info.resource.scaling_group_name,
            target_sgroup_names=info.resource.target_sgroup_names,
            domain_name=info.metadata.domain_name,
            group_id=info.metadata.group_id,
            user_uuid=info.metadata.user_uuid,
            access_key=info.metadata.access_key,
            images=info.image.images,
            tag=info.image.tag or info.metadata.tag,
            occupying_slots=info.resource.occupying_slots,
            requested_slots=info.resource.requested_slots,
            vfolder_mounts=info.mounts.vfolder_mounts,
            environ=info.execution.environ,
            bootstrap_script=info.execution.bootstrap_script,
            startup_command=info.execution.startup_command,
            use_host_network=info.execution.use_host_network,
            batch_timeout=info.lifecycle.batch_timeout,
            created_at=info.lifecycle.created_at
            or info.metadata.created_at
            or datetime.now(tzutc()),
            terminated_at=info.lifecycle.terminated_at,
            starts_at=info.lifecycle.starts_at,
            status=info.lifecycle.status or SessionStatus.PENDING,
            status_info=info.lifecycle.status_info,
            status_data=info.lifecycle.status_data,
            status_history=info.lifecycle.status_history,
            callback_url=info.execution.callback_url,
            result=info.lifecycle.result,
            num_queries=info.metrics.num_queries,
            last_stat=info.metrics.last_stat,
            network_type=info.network.network_type,
            network_id=info.network.network_id,
        )

    def to_session_info(self) -> SessionInfo:
        return SessionInfo(
            identity=SessionIdentity(
                id=self.id,
                creation_id=self.creation_id,
                name=self.name,
                session_type=self.session_type,
                priority=self.priority,
            ),
            metadata=SessionMetadata(
                name=self.name,
                domain_name=self.domain_name,
                group_id=self.group_id,
                user_uuid=self.user_uuid,
                access_key=self.access_key,
                session_type=self.session_type,
                priority=self.priority,
                created_at=self.created_at,
                tag=self.tag,
            ),
            resource=ResourceSpec(
                cluster_mode=self.cluster_mode,
                cluster_size=self.cluster_size,
                occupying_slots=self.occupying_slots,
                requested_slots=self.requested_slots,
                scaling_group_name=self.scaling_group_name,
                target_sgroup_names=self.target_sgroup_names,
                agent_ids=self.agent_ids,
            ),
            image=ImageSpec(
                images=self.images,
                tag=self.tag,
            ),
            mounts=MountSpec(
                vfolder_mounts=self.vfolder_mounts,
            ),
            execution=SessionExecution(
                environ=self.environ,
                bootstrap_script=self.bootstrap_script,
                startup_command=self.startup_command,
                use_host_network=self.use_host_network,
                callback_url=self.callback_url,
            ),
            lifecycle=SessionLifecycle(
                status=self.status,
                result=self.result,
                created_at=self.created_at,
                terminated_at=self.terminated_at,
                starts_at=self.starts_at,
                status_changed=self.status_changed,
                batch_timeout=self.batch_timeout,
                status_info=self.status_info,
                status_data=self.status_data,
                status_history=self.status_history,
            ),
            metrics=SessionMetrics(
                num_queries=self.num_queries,
                last_stat=self.last_stat,
            ),
            network=SessionNetwork(
                network_type=self.network_type,
                network_id=self.network_id,
            ),
        )

    @property
    def vfolders_sorted_by_id(self) -> list[VFolderMount]:
        # TODO: Remove this after ComputeSessionNode and ComputeSession deprecates vfolder_mounts field
        return sorted(self.vfolder_mounts, key=lambda row: row.vfid.folder_id)

    @property
    def main_kernel(self) -> KernelRow:
        kerns = tuple(kern for kern in self.kernels if kern.cluster_role == DEFAULT_ROLE)
        if len(kerns) > 1:
            raise TooManyKernelsFound(
                f"Session (id: {self.id}) has more than 1 main kernel.",
            )
        if len(kerns) == 0:
            raise MainKernelNotFound(
                f"Session (id: {self.id}) has no main kernel.",
            )
        return kerns[0]

    @property
    def status_changed(self) -> Optional[datetime]:
        if self.status_history is None:
            return None
        try:
            return datetime.fromisoformat(self.status_history[self.status.name])
        except KeyError:
            return None

    @property
    def resource_opts(self) -> dict[str, Any]:
        return {kern.cluster_hostname: kern.resource_opts for kern in self.kernels}

    @property
    def is_private(self) -> bool:
        return self.session_type in PRIVATE_SESSION_TYPES

    def get_kernel_by_id(self, kernel_id: KernelId) -> KernelRow:
        kerns = tuple(kern for kern in self.kernels if kern.id == kernel_id)
        if len(kerns) > 1:
            raise TooManyKernelsFound(f"Multiple kernels found (id:{kernel_id}).")
        if len(kerns) == 0:
            raise KernelNotFound(f"Session has no such kernel (sid:{self.id}, kid:{kernel_id}))")
        return kerns[0]

    def get_kernel_by_cluster_name(self, cluster_name: str) -> KernelRow:
        kerns = tuple(kern for kern in self.kernels if kern.cluster_name == cluster_name)
        if len(kerns) > 1:
            raise TooManyKernelsFound(
                f"Session (id: {self.id}) has more than 1 kernel with {cluster_name = }",
            )
        if len(kerns) == 0:
            raise MainKernelNotFound(
                f"Session (id: {self.id}) has no kernel with {cluster_name = }.",
            )
        return kerns[0]

    @classmethod
    async def get_session_id_by_kernel(
        cls, db: ExtendedAsyncSAEngine, kernel_id: KernelId
    ) -> SessionId:
        query = sa.select(KernelRow.session_id).where(KernelRow.id == kernel_id)
        async with db.begin_readonly_session() as db_session:
            return await db_session.scalar(query)

    @classmethod
    async def get_sessions_by_status(
        cls,
        db_session: SASession,
        status: SessionStatus,
        *,
        load_kernel_image: bool = False,
    ) -> list[SessionRow]:
        load_options = selectinload(SessionRow.kernels)
        if load_kernel_image:
            load_options = load_options.options(
                joinedload(KernelRow.image_row).options(joinedload(ImageRow.registry_row))
            )
        stmt = sa.select(SessionRow).where(SessionRow.status == status).options(load_options)
        return (await db_session.scalars(stmt)).all()

    @classmethod
    async def get_session_to_determine_status(
        cls,
        db_session: SASession,
        session_id: SessionId,
    ) -> SessionRow:
        stmt = (
            sa.select(SessionRow)
            .where(SessionRow.id == session_id)
            # TODO: Add kernel loading strategy?
            .options(selectinload(SessionRow.kernels))
        )
        session_row = cast(SessionRow | None, await db_session.scalar(stmt))
        if session_row is None:
            raise SessionNotFound(f"Session not found (id:{session_id})")
        return session_row

    def determine_and_set_status(
        self,
        status_info: str | None = None,
        status_data: Mapping[str, Any] | JSONCoalesceExpr | None = None,
        status_changed_at: datetime | None = None,
    ) -> bool:
        """
        Determine the current status of a session based on its sibling kernels.
        If it is possible to transit from the current status to the determined status, set status.
        Else, do nothing.
        Return True if a transition happened, else return False.
        """
        determined_status = determine_session_status_by_kernels(self.kernels)
        if determined_status not in SESSION_STATUS_TRANSITION_MAP[self.status]:
            return False

        self.set_status(determined_status, status_info, status_data, status_changed_at)
        return True

    @classmethod
    async def list_session_by_condition(
        cls,
        conditions: Iterable[QueryCondition],
        options: Iterable[QueryOption] = tuple(),
        *,
        db: ExtendedAsyncSAEngine,
    ) -> list[Self]:
        stmt = sa.select(SessionRow)
        for cond in conditions:
            stmt = cond(stmt)
        for option in options:
            stmt = option(stmt)

        async def fetch(db_session: SASession) -> list[SessionRow]:
            return (await db_session.scalars(stmt)).all()

        async with db.connect() as db_conn:
            return await execute_with_txn_retry(fetch, db.begin_readonly_session, db_conn)

    def set_status(
        self,
        status: SessionStatus,
        status_info: str | None = None,
        status_data: Mapping[str, Any] | JSONCoalesceExpr | None = None,
        status_changed_at: datetime | None = None,
    ) -> None:
        """
        Set the status of the session.
        """
        now = status_changed_at or datetime.now(tzutc())
        if status in (SessionStatus.CANCELLED, SessionStatus.TERMINATED):
            self.terminated_at = now
        self.status = status
        self.status_history = {
            **self.status_history,
            status.name: now.isoformat(),
        }
        if status_data is not None:
            self.status_data = status_data

        _status_info: str | None = None
        if status_info is None:
            _status_info = self.main_kernel.status_info
        else:
            _status_info = status_info
        if _status_info is not None:
            self.status_info = _status_info

    def delegate_ownership(self, user_uuid: UUID, access_key: AccessKey) -> None:
        self.user_uuid = user_uuid
        self.access_key = access_key
        for kernel_row in cast(list[KernelRow], self.kernels):
            kernel_row.delegate_ownership(user_uuid, access_key)

    @staticmethod
    async def delete_by_user_id(user_uuid: UUID, *, db_session: SASession) -> None:
        await db_session.execute(sa.delete(KernelRow).where(KernelRow.user_uuid == user_uuid))
        await db_session.execute(sa.delete(SessionRow).where(SessionRow.user_uuid == user_uuid))

    @staticmethod
    async def set_session_status(
        db: ExtendedAsyncSAEngine,
        session_id: SessionId,
        status: SessionStatus,
        *,
        status_data: Optional[Mapping[str, Any]] = None,
        reason: Optional[str] = None,
        status_changed_at: Optional[datetime] = None,
    ) -> None:
        if status_changed_at is None:
            now = datetime.now(tzutc())
        else:
            now = status_changed_at
        data = {
            "status": status,
            "status_history": sql_json_merge(
                SessionRow.status_history,
                (),
                {
                    status.name: datetime.now(tzutc()).isoformat(),
                },
            ),
        }
        if status_data is not None:
            data["status_data"] = status_data
        if reason is not None:
            data["status_info"] = reason
        if status in (SessionStatus.CANCELLED, SessionStatus.TERMINATED):
            data["terminated_at"] = now

        async def _update() -> None:
            async with db.begin_session() as db_sess:
                query = sa.update(SessionRow).values(**data).where(SessionRow.id == session_id)
                await db_sess.execute(query)

        await execute_with_retry(_update)

    @classmethod
    async def set_session_result(
        cls,
        db: ExtendedAsyncSAEngine,
        session_id: SessionId,
        success: bool,
        exit_code: int,
    ) -> None:
        # TODO: store exit code?
        data = {
            "result": SessionResult.SUCCESS if success else SessionResult.FAILURE,
        }

        async def _update() -> None:
            async with db.begin_session() as db_sess:
                query = sa.update(SessionRow).values(**data).where(SessionRow.id == session_id)
                await db_sess.execute(query)

        await execute_with_retry(_update)

    @classmethod
    async def match_sessions(
        cls,
        db_session: SASession,
        session_reference: str | UUID | list[UUID],
        access_key: Optional[AccessKey],
        *,
        allow_prefix: bool = False,
        allow_stale: bool = True,
        for_update: bool = False,
        max_matches: Optional[int] = 10,
        eager_loading_op: Optional[Sequence] = None,
    ) -> List[SessionRow]:
        """
        Match the prefix of session ID or session name among the sessions
        that belongs to the given access key, and return the list of SessionRow.
        """

        if isinstance(session_reference, list):
            query_list = [
                aiotools.apartial(
                    _match_sessions_by_id,
                    session_id_or_list=session_reference,
                    allow_prefix=False,
                )
            ]
        else:
            query_list = [
                aiotools.apartial(
                    _match_sessions_by_name,
                    session_name=str(session_reference),
                    allow_prefix=allow_prefix,
                )
            ]
            try:
                session_id = UUID(str(session_reference))
                # Fetch id-based query first
                query_list = [
                    aiotools.apartial(
                        _match_sessions_by_id,
                        session_id_or_list=SessionId(session_id),
                        allow_prefix=False,
                    ),
                    *query_list,
                ]
                if allow_prefix:
                    query_list = [
                        aiotools.apartial(
                            _match_sessions_by_id,
                            session_id_or_list=SessionId(session_id),
                            allow_prefix=True,
                        ),
                        *query_list,
                    ]
            except ValueError:
                pass

        for fetch_func in query_list:
            rows = await fetch_func(
                db_session,
                access_key=access_key,
                allow_stale=allow_stale,
                for_update=for_update,
                max_matches=max_matches,
                eager_loading_op=eager_loading_op,
            )
            if not rows:
                continue
            return rows
        return []

    @classmethod
    async def get_session(
        cls,
        db_session: SASession,
        session_name_or_id: Union[str, UUID],
        access_key: Optional[AccessKey] = None,
        *,
        allow_stale: bool = False,
        for_update: bool = False,
        kernel_loading_strategy: KernelLoadingStrategy = KernelLoadingStrategy.NONE,
        eager_loading_op: list[Any] | None = None,
    ) -> SessionRow:
        """
        Retrieve the session information by session's UUID,
        or session's name paired with access_key.
        This will return the information of the session and the sibling kernel(s).

        :param db_session: Database connection to use when fetching row.
        :param session_name_or_id: Name or ID (UUID) of session to look up.
        :param access_key: Access key used to create session.
        :param allow_stale: If set to True, filter "inactive" sessions as well as "active" ones.
                            Otherwise filter "active" sessions only.
        :param for_update: Apply for_update during executing select query.
        :param kernel_loading_strategy: Determines JOIN strategy of `kernels` relation when fetching session rows.
        :param eager_loading_op: Extra loading operators to be passed directly to `match_sessions()` API.
        """
        _eager_loading_op = eager_loading_op or []
        match kernel_loading_strategy:
            case KernelLoadingStrategy.ALL_KERNELS:
                _eager_loading_op.extend([
                    noload("*"),
                    selectinload(SessionRow.kernels).options(
                        noload("*"),
                        selectinload(KernelRow.agent_row).noload("*"),
                    ),
                ])
            case KernelLoadingStrategy.MAIN_KERNEL_ONLY:
                kernel_rel = SessionRow.kernels
                kernel_rel.and_(KernelRow.cluster_role == DEFAULT_ROLE)
                _eager_loading_op.extend([
                    noload("*"),
                    selectinload(kernel_rel).options(
                        noload("*"),
                        selectinload(KernelRow.agent_row).noload("*"),
                    ),
                ])
        _eager_loading_op.append(joinedload(SessionRow.user))

        session_list = await cls.match_sessions(
            db_session,
            session_name_or_id,
            access_key,
            allow_stale=allow_stale,
            for_update=for_update,
            eager_loading_op=_eager_loading_op,
        )
        if not session_list:
            raise SessionNotFound(f"Session (id={session_name_or_id}) does not exist.")
        if len(session_list) > 1:
            session_infos = [
                {
                    "session_id": sess.id,
                    "session_name": sess.name,
                    "status": sess.status,
                    "created_at": sess.created_at,
                }
                for sess in session_list
            ]
            raise TooManySessionsMatched(extra_data={"matches": session_infos})
        return session_list[0]

    @classmethod
    async def list_sessions(
        cls,
        db_session: SASession,
        session_ids: list[UUID],
        access_key: Optional[AccessKey] = None,
        *,
        allow_stale: bool = False,
        for_update: bool = False,
        kernel_loading_strategy=KernelLoadingStrategy.NONE,
        eager_loading_op: list[Any] | None = None,
        max_load_count: Optional[int] = None,
    ) -> Iterable[SessionRow]:
        _eager_loading_op = eager_loading_op or []
        match kernel_loading_strategy:
            case KernelLoadingStrategy.ALL_KERNELS:
                _eager_loading_op.extend([
                    noload("*"),
                    selectinload(SessionRow.kernels).options(
                        noload("*"),
                        selectinload(KernelRow.agent_row).noload("*"),
                    ),
                ])
            case KernelLoadingStrategy.MAIN_KERNEL_ONLY:
                kernel_rel = SessionRow.kernels
                kernel_rel.and_(KernelRow.cluster_role == DEFAULT_ROLE)
                _eager_loading_op.extend([
                    noload("*"),
                    selectinload(kernel_rel).options(
                        noload("*"),
                        selectinload(KernelRow.agent_row).noload("*"),
                    ),
                ])

        session_list = await cls.match_sessions(
            db_session,
            session_ids,
            access_key,
            allow_stale=allow_stale,
            for_update=for_update,
            eager_loading_op=_eager_loading_op,
            max_matches=max_load_count,
        )
        try:
            return session_list
        except IndexError:
            raise SessionNotFound(f"Session (ids={session_ids}) does not exist.")

    @classmethod
    async def get_session_by_id(
        cls,
        db_session: SASession,
        session_id: SessionId,
        access_key: Optional[AccessKey] = None,
        *,
        max_matches: int | None = None,
        allow_stale: bool = True,
        for_update: bool = False,
        eager_loading_op=None,
    ) -> SessionRow:
        sessions = await _match_sessions_by_id(
            db_session,
            session_id,
            access_key,
            max_matches=max_matches,
            allow_stale=allow_stale,
            for_update=for_update,
            eager_loading_op=eager_loading_op,
            allow_prefix=False,
        )
        try:
            return sessions[0]
        except IndexError:
            raise SessionNotFound(f"Session (id={session_id}) does not exist.")

    @classmethod
    async def get_sgroup_managed_sessions(
        cls,
        db_sess: SASession,
        sgroup_name: str,
    ) -> List[SessionRow]:
        candidate_statues = (SessionStatus.PENDING, *AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES)
        query = (
            sa.select(SessionRow)
            .where(
                (SessionRow.scaling_group_name == sgroup_name)
                & (SessionRow.status.in_(candidate_statues))
            )
            .options(
                noload("*"),
                selectinload(SessionRow.group).options(noload("*")),
                selectinload(SessionRow.domain).options(noload("*")),
                selectinload(SessionRow.access_key_row).options(noload("*")),
                selectinload(SessionRow.kernels).options(noload("*")),
            )
        )
        result = await db_sess.execute(query)
        return result.scalars().all()

    async def get_network_ref(self, db_sess: SASession) -> str | None:
        if not self.network_id or not self.network_type:
            return None
        match self.network_type:
            case NetworkType.VOLATILE | NetworkType.HOST:
                return self.network_id
            case NetworkType.PERSISTENT:
                network_row = await NetworkRow.get(db_sess, UUID(self.network_id))
                return network_row.ref_name
            case _:
                return None

    @classmethod
    def get_status_elapsed_time(
        cls, status: SessionStatus, until: datetime
    ) -> sa.sql.elements.BinaryExpression:
        return until - cls.status_history[status.name].astext.cast(sa.types.DateTime(timezone=True))


def by_status(statuses: Iterable[SessionStatus]) -> QueryCondition:
    def _by_status(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(SessionRow.status.in_(statuses))

    return _by_status


def by_user_id(user_id: UUID) -> QueryCondition:
    def _by_user_id(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(SessionRow.user_uuid == user_id)

    return _by_user_id


def by_project_id(project_id: UUID) -> QueryCondition:
    def _by_project_id(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(SessionRow.group_id == project_id)

    return _by_project_id


def by_domain_name(domain_name: str) -> QueryCondition:
    def _by_domain_name(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(SessionRow.domain_name == domain_name)

    return _by_domain_name


def by_resource_group_name(resource_group_name: str) -> QueryCondition:
    def _by_resource_group_name(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(SessionRow.scaling_group_name == resource_group_name)

    return _by_resource_group_name


def by_raw_filter(filter_spec: FieldSpecType, raw_filter: str) -> QueryCondition:
    def _by_raw_filter(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        qfparser = QueryFilterParser(filter_spec)
        new_cond = qfparser.parse_filter(SessionRow, raw_filter)
        return query_stmt.where(new_cond)

    return _by_raw_filter


class SessionLifecycleManager:
    status_set_key = "session_status_update"

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_stat: ValkeyStatClient,
        valkey_live: ValkeyLiveClient,
        event_producer: EventProducer,
        hook_plugin_ctx: HookPluginContext,
        registry: AgentRegistry,
    ) -> None:
        self.db = db
        self.valkey_stat = valkey_stat
        self.valkey_live = valkey_live
        self.event_producer = event_producer
        self.hook_plugin_ctx = hook_plugin_ctx
        self.registry = registry

        def _encode(sid: SessionId) -> bytes:
            return sid.bytes

        def _decode(raw_sid: bytes) -> SessionId:
            return SessionId(UUID(bytes=raw_sid))

        self._encoder = _encode
        self._decoder = _decode

    async def _transit_session_status(
        self,
        db_conn: SAConnection,
        session_id: SessionId,
        status_changed_at: datetime | None = None,
    ) -> tuple[SessionRow, bool]:
        now = status_changed_at or datetime.now(tzutc())

        async def _get_and_transit(
            db_session: SASession,
        ) -> tuple[SessionRow, bool]:
            session_row = await SessionRow.get_session_to_determine_status(db_session, session_id)
            transited = session_row.determine_and_set_status(status_changed_at=now)

            def _calculate_session_occupied_slots(session_row: SessionRow):
                session_occupying_slots = ResourceSlot()
                for row in session_row.kernels:
                    kernel_row = cast(KernelRow, row)
                    kernel_allocs = kernel_row.occupied_slots
                    session_occupying_slots.sync_keys(kernel_allocs)
                    for key, val in session_occupying_slots.items():
                        session_occupying_slots[key] = str(
                            Decimal(val) + Decimal(kernel_allocs[key])
                        )
                session_row.occupying_slots = session_occupying_slots

            match session_row.status:
                case SessionStatus.CREATING:
                    _calculate_session_occupied_slots(session_row)
                case SessionStatus.RUNNING if transited:
                    _calculate_session_occupied_slots(session_row)

            return session_row, transited

        return await execute_with_txn_retry(_get_and_transit, self.db.begin_session, db_conn)

    async def _post_status_transition(
        self,
        session_row: SessionRow,
    ) -> None:
        match session_row.status:
            case SessionStatus.PREPARED:
                await self.event_producer.anycast_event(DoStartSessionEvent())
            case SessionStatus.RUNNING:
                log.debug(
                    "Producing SessionStartedEvent({}, {})",
                    session_row.id,
                    session_row.creation_id,
                )
                await self.event_producer.anycast_and_broadcast_event(
                    SessionStartedAnycastEvent(session_row.id, session_row.creation_id),
                    SessionStartedBroadcastEvent(session_row.id, session_row.creation_id),
                )
                await self.hook_plugin_ctx.notify(
                    "POST_START_SESSION",
                    (
                        session_row.id,
                        session_row.name,
                        session_row.access_key,
                    ),
                )
                match session_row.session_type:
                    case SessionTypes.BATCH:
                        await self.registry.trigger_batch_execution(session_row)
                    case SessionTypes.INFERENCE:
                        await self.handle_inference_session_update(session_row)
            case SessionStatus.TERMINATING:
                if session_row.session_type == SessionTypes.INFERENCE:
                    async with self.db.begin_session() as db_sess:
                        route = await RoutingRow.get_by_session(db_sess, session_row.id)
                        route.status = RouteStatus.TERMINATING
                        await db_sess.commit()
                    await self.handle_inference_session_update(session_row)
            case SessionStatus.TERMINATED:
                if session_row.session_type == SessionTypes.INFERENCE:
                    async with self.db.begin_session() as db_sess:
                        query = sa.delete(RoutingRow).where(RoutingRow.session == session_row.id)
                        await db_sess.execute(query)
                        await db_sess.commit()
                await self.event_producer.anycast_and_broadcast_event(
                    SessionTerminatedAnycastEvent(
                        session_row.id, session_row.main_kernel.status_info
                    ),
                    SessionTerminatedBroadcastEvent(
                        session_row.id, session_row.main_kernel.status_info
                    ),
                )
            case _:
                pass

    async def handle_inference_session_update(self, session: SessionRow) -> None:
        async with self.db.begin_readonly_session() as db_sess:
            route = await RoutingRow.get_by_session(db_sess, session.id, load_endpoint=True)
        await self.registry.notify_endpoint_route_update_to_appproxy(route.endpoint)

    async def transit_session_status(
        self,
        session_ids: Iterable[SessionId],
        status_changed_at: datetime | None = None,
    ) -> list[tuple[SessionRow, bool]]:
        if not session_ids:
            return []
        now = status_changed_at or datetime.now(tzutc())

        async def _transit(_db_conn: SAConnection) -> list[tuple[SessionRow, bool]]:
            result: list[tuple[SessionRow, bool]] = []
            for sid in session_ids:
                row, is_transited = await self._transit_session_status(_db_conn, sid, now)
                result.append((row, is_transited))
            return result

        async with self.db.connect() as db_conn:
            result = await _transit(db_conn)
        for session_row, is_transited in result:
            if is_transited:
                await self._post_status_transition(session_row)
        return result

    async def register_status_updatable_session(self, session_ids: Iterable[SessionId]) -> None:
        if not session_ids:
            return

        try:
            await self.valkey_stat.register_session_ids_for_status_update(
                self.status_set_key,
                [self._encoder(sid) for sid in session_ids],
            )
        except (
            redis.exceptions.RedisError,
            redis.exceptions.RedisClusterException,
            redis.exceptions.ChildDeadlockedError,
        ) as e:
            log.warning(f"Failed to update session status to redis, skip. (e:{repr(e)})")
        await self.event_producer.anycast_event(DoUpdateSessionStatusEvent())

    async def get_status_updatable_sessions(self) -> set[SessionId]:
        try:
            results = await self.valkey_stat.get_and_clear_session_ids_for_status_update(
                self.status_set_key,
            )
        except (
            redis.exceptions.RedisError,
            redis.exceptions.RedisClusterException,
            redis.exceptions.ChildDeadlockedError,
        ) as e:
            log.warning(f"Failed to fetch session status data from redis, skip. (e:{repr(e)})")
            results = []
        result: list[SessionId] = []
        for raw_session_id in results:
            try:
                result.append(self._decoder(raw_session_id))
            except (ValueError, SyntaxError):
                log.warning(f"Cannot parse session id, skip. (id:{raw_session_id!r})")
                continue

        async with self.db.begin_readonly_session() as db_session:
            session_query = sa.select(SessionRow).where(
                SessionRow.status.in_(SessionStatus.kernel_awaiting_statuses())
            )
            session_rows = await db_session.scalars(session_query)
            session_ids = [row.id for row in session_rows]
        return {*result, *session_ids}

    async def deregister_status_updatable_session(
        self,
        session_ids: Iterable[SessionId],
    ) -> int:
        if not session_ids:
            return 0

        try:
            ret = await self.valkey_stat.remove_session_ids_from_status_update(
                self.status_set_key,
                [self._encoder(sid) for sid in session_ids],
            )
        except (
            redis.exceptions.RedisError,
            redis.exceptions.RedisClusterException,
            redis.exceptions.ChildDeadlockedError,
        ) as e:
            log.warning(f"Failed to remove session status data from redis, skip. (e:{repr(e)})")
            return 0
        return ret


class SessionDependencyRow(Base):
    __tablename__ = "session_dependencies"
    session_id = sa.Column(
        "session_id",
        GUID,
        sa.ForeignKey("sessions.id", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    depends_on = sa.Column(
        "depends_on",
        GUID,
        sa.ForeignKey("sessions.id", onupdate="CASCADE", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )

    __table_args__ = (
        # constraint
        sa.PrimaryKeyConstraint("session_id", "depends_on", name="sess_dep_pk"),
    )


async def check_all_dependencies(
    db_session: SASession,
    sess_ctx: SessionRow,
) -> List[SessionRow]:
    j = sa.join(
        SessionDependencyRow,
        SessionRow,
        SessionDependencyRow.depends_on == SessionDependencyRow.session_id,
    )
    query = (
        sa.select(SessionRow.id, SessionRow.name, SessionRow.result)
        .select_from(j)
        .where(SessionDependencyRow.session_id == sess_ctx.id)
    )
    result = await db_session.execute(query)
    rows = result.scalars().all()
    pending_dependencies = [
        sess_row for sess_row in rows if SessionResult(sess_row.result) != SessionResult.SUCCESS
    ]
    return pending_dependencies


DEFAULT_SESSION_ORDERING = [
    sa.desc(
        sa.func.greatest(
            SessionRow.created_at,
            SessionRow.terminated_at,
        )
    ),
]


ALL_COMPUTE_SESSION_PERMISSIONS: frozenset[ComputeSessionPermission] = frozenset([
    perm for perm in ComputeSessionPermission
])
OWNER_PERMISSIONS: frozenset[ComputeSessionPermission] = ALL_COMPUTE_SESSION_PERMISSIONS
ADMIN_PERMISSIONS: frozenset[ComputeSessionPermission] = ALL_COMPUTE_SESSION_PERMISSIONS
MONITOR_PERMISSIONS: frozenset[ComputeSessionPermission] = frozenset({
    ComputeSessionPermission.READ_ATTRIBUTE,
    ComputeSessionPermission.UPDATE_ATTRIBUTE,
})
PRIVILEGED_MEMBER_PERMISSIONS: frozenset[ComputeSessionPermission] = frozenset()
MEMBER_PERMISSIONS: frozenset[ComputeSessionPermission] = frozenset()

WhereClauseType: TypeAlias = (
    sa.sql.expression.BinaryExpression | sa.sql.expression.BooleanClauseList
)


class ComputeSessionPermissionContext(
    AbstractPermissionContext[ComputeSessionPermission, SessionRow, SessionId]
):
    @property
    def query_condition(self) -> WhereClauseType | None:
        cond: WhereClauseType | None = None

        def _OR_coalesce(
            base_cond: WhereClauseType | None,
            _cond: sa.sql.expression.BinaryExpression,
        ) -> WhereClauseType:
            return base_cond | _cond if base_cond is not None else _cond

        if self.user_id_to_permission_map:
            cond = _OR_coalesce(
                cond, SessionRow.user_uuid.in_(self.user_id_to_permission_map.keys())
            )
        if self.project_id_to_permission_map:
            cond = _OR_coalesce(
                cond, SessionRow.group_id.in_(self.project_id_to_permission_map.keys())
            )
        if self.domain_name_to_permission_map:
            cond = _OR_coalesce(
                cond, SessionRow.domain_name.in_(self.domain_name_to_permission_map.keys())
            )
        if self.object_id_to_additional_permission_map:
            cond = _OR_coalesce(
                cond, SessionRow.id.in_(self.object_id_to_additional_permission_map.keys())
            )
        if self.object_id_to_overriding_permission_map:
            cond = _OR_coalesce(
                cond, SessionRow.id.in_(self.object_id_to_overriding_permission_map.keys())
            )
        return cond

    async def build_query(self) -> sa.sql.Select | None:
        cond = self.query_condition
        if cond is None:
            return None
        return sa.select(SessionRow).where(cond)

    async def calculate_final_permission(
        self, rbac_obj: SessionRow
    ) -> frozenset[ComputeSessionPermission]:
        session_row = rbac_obj
        session_id = cast(SessionId, session_row.id)
        permissions: frozenset[ComputeSessionPermission] = frozenset()

        if (
            overriding_perm := self.object_id_to_overriding_permission_map.get(session_id)
        ) is not None:
            permissions = overriding_perm
        else:
            permissions |= self.object_id_to_additional_permission_map.get(session_id, set())
            permissions |= self.user_id_to_permission_map.get(session_row.user_uuid, set())
            permissions |= self.project_id_to_permission_map.get(session_row.group_id, set())
            permissions |= self.domain_name_to_permission_map.get(session_row.domain_name, set())
        return permissions


class ComputeSessionPermissionContextBuilder(
    AbstractPermissionContextBuilder[ComputeSessionPermission, ComputeSessionPermissionContext]
):
    db_session: SASession

    def __init__(self, db_session: SASession) -> None:
        self.db_session = db_session

    @override
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[ComputeSessionPermission]:
        roles = await get_predefined_roles_in_scope(ctx, target_scope, self.db_session)
        permissions = await self._calculate_permission_by_predefined_roles(roles)
        return permissions

    @override
    async def build_ctx_in_system_scope(
        self,
        ctx: ClientContext,
    ) -> ComputeSessionPermissionContext:
        from .domain import DomainRow

        perm_ctx = ComputeSessionPermissionContext()
        _domain_query_stmt = sa.select(DomainRow).options(load_only(DomainRow.name))
        for row in await self.db_session.scalars(_domain_query_stmt):
            to_be_merged = await self.build_ctx_in_domain_scope(ctx, DomainScope(row.name))
            perm_ctx.merge(to_be_merged)
        return perm_ctx

    @override
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> ComputeSessionPermissionContext:
        permission_ctx = await self._build_at_domain_scope_non_recursively(ctx, scope.domain_name)
        _user_perm_ctx = await self._build_at_user_scope_in_domain(
            ctx, ctx.user_id, scope.domain_name
        )
        permission_ctx.merge(_user_perm_ctx)
        _project_perm_ctx = await self._build_at_project_scopes_in_domain(ctx, scope.domain_name)
        permission_ctx.merge(_project_perm_ctx)
        return permission_ctx

    @override
    async def build_ctx_in_project_scope(
        self,
        ctx: ClientContext,
        scope: ProjectScope,
    ) -> ComputeSessionPermissionContext:
        permission_ctx = await self._build_at_project_scope_non_recursively(ctx, scope.project_id)
        _user_perm_ctx = await self._build_at_user_scope_in_project(
            ctx, ctx.user_id, scope.project_id
        )
        permission_ctx.merge(_user_perm_ctx)
        return permission_ctx

    @override
    async def build_ctx_in_user_scope(
        self,
        ctx: ClientContext,
        scope: UserRBACScope,
    ) -> ComputeSessionPermissionContext:
        return await self._build_at_user_scope_non_recursively(ctx, scope.user_id)

    async def _build_at_domain_scope_non_recursively(
        self,
        ctx: ClientContext,
        domain_name: str,
    ) -> ComputeSessionPermissionContext:
        permissions = await self.calculate_permission(ctx, DomainScope(domain_name))
        result = ComputeSessionPermissionContext(
            domain_name_to_permission_map={domain_name: permissions}
        )
        return result

    async def _build_at_user_scope_in_domain(
        self,
        ctx: ClientContext,
        user_id: UUID,
        domain_name: str,
    ) -> ComputeSessionPermissionContext:
        # For Superadmin and monitor who can create objects in multiple different domains.
        permissions = await self.calculate_permission(ctx, UserRBACScope(user_id, domain_name))

        _vfolder_stmt = (
            sa.select(SessionRow)
            .where((SessionRow.user_uuid == user_id) & (SessionRow.domain_name == domain_name))
            .options(load_only(SessionRow.id))
        )
        own_folder_map = {
            row.id: permissions for row in await self.db_session.scalars(_vfolder_stmt)
        }
        result = ComputeSessionPermissionContext(
            object_id_to_additional_permission_map=own_folder_map
        )
        return result

    async def _build_at_user_scope_in_project(
        self,
        ctx: ClientContext,
        user_id: UUID,
        project_id: UUID,
    ) -> ComputeSessionPermissionContext:
        permissions = await self.calculate_permission(ctx, UserRBACScope(user_id))

        _vfolder_stmt = (
            sa.select(SessionRow)
            .where((SessionRow.user_uuid == user_id) & (SessionRow.group_id == project_id))
            .options(load_only(SessionRow.id))
        )
        own_folder_map = {
            row.id: permissions for row in await self.db_session.scalars(_vfolder_stmt)
        }
        result = ComputeSessionPermissionContext(
            object_id_to_additional_permission_map=own_folder_map
        )
        return result

    async def _build_at_project_scopes_in_domain(
        self,
        ctx: ClientContext,
        domain_name: str,
    ) -> ComputeSessionPermissionContext:
        result = ComputeSessionPermissionContext()

        _project_stmt = (
            sa.select(GroupRow)
            .where(GroupRow.domain_name == domain_name)
            .options(load_only(GroupRow.id))
        )
        for row in await self.db_session.scalars(_project_stmt):
            _row = cast(GroupRow, row)
            _project_perm_ctx = await self._build_at_project_scope_non_recursively(ctx, _row.id)
            result.merge(_project_perm_ctx)
        return result

    async def _build_at_project_scope_non_recursively(
        self,
        ctx: ClientContext,
        project_id: UUID,
    ) -> ComputeSessionPermissionContext:
        permissions = await self.calculate_permission(ctx, ProjectScope(project_id))
        result = ComputeSessionPermissionContext(
            project_id_to_permission_map={project_id: permissions}
        )
        return result

    async def _build_at_user_scope_non_recursively(
        self,
        ctx: ClientContext,
        user_id: UUID,
    ) -> ComputeSessionPermissionContext:
        permissions = await self.calculate_permission(ctx, UserRBACScope(user_id))
        result = ComputeSessionPermissionContext(user_id_to_permission_map={user_id: permissions})
        return result

    @override
    @classmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[ComputeSessionPermission]:
        return OWNER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[ComputeSessionPermission]:
        return ADMIN_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[ComputeSessionPermission]:
        return MONITOR_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[ComputeSessionPermission]:
        return PRIVILEGED_MEMBER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[ComputeSessionPermission]:
        return MEMBER_PERMISSIONS


async def get_permission_ctx(
    db_conn: SAConnection,
    ctx: ClientContext,
    target_scope: ScopeType,
    requested_permission: ComputeSessionPermission,
) -> ComputeSessionPermissionContext:
    async with ctx.db.begin_readonly_session(db_conn) as db_session:
        builder = ComputeSessionPermissionContextBuilder(db_session)
        permission_ctx = await builder.build(ctx, target_scope, requested_permission)
    return permission_ctx



"""Update Endpoint and Routing for sokovan

Revision ID: fb01c44db0d6
Revises: 43c7e60c56af
Create Date: 2025-08-29 17:47:24.086038

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fb01c44db0d6"
down_revision = "43c7e60c56af"
branch_labels = None
depends_on = None

endpoint_lifecycle_enum_name = "endpointlifecycle"
PENDING = "pending"
SCALING = "scaling"
READY = "ready"
route_status_enum_name = "routestatus"
TERMINATED = "terminated"


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "endpoints",
        sa.Column("desired_replicas", sa.Integer(), server_default=sa.text("NULL"), nullable=True),
    )
    op.add_column("routings", sa.Column("weight", sa.Integer(), nullable=True))
    op.execute(f"ALTER TYPE {endpoint_lifecycle_enum_name} ADD VALUE '{PENDING}'")
    op.execute(f"ALTER TYPE {endpoint_lifecycle_enum_name} ADD VALUE '{SCALING}'")
    op.execute(f"ALTER TYPE {endpoint_lifecycle_enum_name} ADD VALUE '{READY}'")
    op.execute(f"ALTER TYPE {route_status_enum_name} ADD VALUE '{TERMINATED}'")
    op.create_index(
        "ix_endpoints_unique_name_when_not_destroyed",
        "endpoints",
        ["name", "domain", "project"],
        unique=True,
        postgresql_where=sa.text("lifecycle_stage != 'destroyed'"),
    )
    op.create_index(
        "ix_sessions_unique_name_nonterminal",
        "sessions",
        ["name"],
        unique=True,
        postgresql_where=sa.text("status NOT IN ('ERROR', 'TERMINATED', 'CANCELLED')"),
    )


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "ix_sessions_unique_name_nonterminal",
        table_name="sessions",
        postgresql_where=sa.text("status NOT IN ('ERROR', 'TERMINATED', 'CANCELLED')"),
    )
    op.drop_index(
        "ix_endpoints_unique_name_when_not_destroyed",
        table_name="endpoints",
        postgresql_where=sa.text("lifecycle_stage != 'destroyed'"),
    )
    op.drop_column("endpoints", "desired_replicas")
